""" The LifeCycle is a core component for the internal workings of Markata.  It
sets fourth the hooks available, the methods to run them on the Markata
instance, and the order they run in.

### Usage

``` python
from markata import Lifecycle

step = Lifecycle.glob
```

"""
from enum import Enum, auto
from functools import total_ordering


@total_ordering
class LifeCycle(Enum):
    """
    LifeCycle currently supports the following steps.

    * configure - load and fix configuration
    * glob - find files
    * load - load files
    * pre_render - clean up files/metadata before render
    * render - render content
    * post_render - clean up rendered content
    * save - store results to disk

    """

    configure = auto()
    glob = auto()
    load = auto()
    pre_render = auto()
    render = auto()
    post_render = auto()
    save = auto()

    def __lt__(self, other: object) -> bool:
        """
        Determine whether other is less than this instance.
        """
        if isinstance(other, LifeCycle):
            return self.value < other.value
        if isinstance(other, int):
            return self.value < other
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        """
        Determine whether other is equal to this instance.
        """
        if isinstance(other, LifeCycle):
            return self.value == other.value
        if isinstance(other, int):
            return self.value == other
        return NotImplemented



from pathlib import Path
from typing import TYPE_CHECKING

from jinja2 import Template, Undefined

from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata import Markata


class SilentUndefined(Undefined):
    def _fail_with_undefined_error(self, *args, **kwargs):
        return ""


@hook_impl
def render(markata: "Markata") -> None:
    if "post_template" in markata.config:
        template_file = markata.config["post_template"]
    else:
        template_file = Path(__file__).parent / "default_post_template.html"
    with open(template_file) as f:
        template = Template(f.read(), undefined=SilentUndefined)
    for article in markata.iter_articles("apply template"):

        article.html = template.render(
            body=article.html,
            toc=markata.md.toc,  # type: ignore
            config=markata.config,
            **article,
        )



import shutil
from pathlib import Path
from typing import TYPE_CHECKING

from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata import Markata


@hook_impl
def save(markata: "Markata") -> None:
    try:
        output_dir = Path(str(markata.config["output_dir"]))
        assets_dir = Path(str(markata.config["assets_dir"]))
    except KeyError:
        return

    with markata.console.status("copying assets", spinner="aesthetic", speed=0.2):
        if assets_dir.exists():
            shutil.copytree(assets_dir, output_dir, dirs_exist_ok=True)



"""
A Markata plugin to create automatic descriptions for markdown documents.  It
does this by grabbing the first `{len}` number of characters from the document
that are in a paragraph.

## Configuration

Open up your `markata.toml` file and add new entries for your
auto_descriptions.  You can have multiple desriptions, each one will be named
after the key you give it in your config.

``` toml
[markata]

# make sure its in your list of hooks
hooks=[
   "markata.plugins.auto_description",
   ]

[markata.auto_description.description]
len=160
[markata.auto_description.long_description]
len=250
[markata.auto_description.super_description]
len=500
```

!!! note
   Make sure that you have the auto_description plugin in your configured hooks.

In the above we will end up with three different descritpions,
(`description`, `long_description`, and `super_description`) each will be the
first number of characters from the document as specified in the config.

### Defaults

By default markata will set `description` to 160 and `long_description` to 250,
if they are not set in your config.

### Using the Description

Downstream hooks can now use the description for things such as seo, or feeds.
Here is a simple example that lists all of the descriptions in all posts.  This
is a handy thing you can do right from a repl.

``` python
from markata import Markata
m = Markata()
[p["description"] for p in m.articles]
```

"""
from itertools import compress
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict

import commonmark

from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from diskcache import FanoutCache
    from frontmatter import Post

    from markata import Markata

_parser = commonmark.Parser()


def get_description(article: "Post") -> str:
    """
    Get the full-length description for a single post using the commonmark
    parser.  Only paragraph nodes will count as text towards the description.
    """
    ast = _parser.parse(article.content)

    # find all paragraph nodes
    paragraph_nodes = [
        n[0]
        for n in ast.walker()
        if n[0].t == "paragraph" and n[0].first_child.literal is not None
    ]
    # for reasons unknown to me commonmark duplicates nodes, dedupe based on sourcepos
    sourcepos = [p.sourcepos for p in paragraph_nodes]
    # find first occurence of node based on source position
    unique_mask = [sourcepos.index(s) == i for i, s in enumerate(sourcepos)]
    # deduplicate paragraph_nodes based on unique source position
    unique_paragraph_nodes = list(compress(paragraph_nodes, unique_mask))
    paragraphs = " ".join([p.first_child.literal for p in unique_paragraph_nodes])
    return paragraphs


def set_description(
    markata: "Markata",
    article: "Post",
    cache: "FanoutCache",
    config: Dict,
    max_description: int = 500,
    plugin_text: None = "",
) -> None:
    """
    For a given `article`, find the description, put it in the cache, and set
    the configured descriptions for the article.
    """
    key = markata.make_hash(
        article.content,
        plugin_text,
        config,
    )

    description_from_cache = cache.get(key)
    if description_from_cache is None:
        description = get_description(article)[:max_description]
        markata.cache.add(key, description, expire=config["cache_expire"])
    else:
        description = description_from_cache

    for description_key in config:
        if description_key not in ["cache_expire", "config_key"]:

            # overwrites missing (None) and empty ('')
            if not article.metadata.get(description_key):
                article.metadata[description_key] = description[
                    : config[description_key]["len"]
                ]


@hook_impl
def pre_render(markata: "Markata") -> None:
    """
    The Markata hook that will set descriptions for all posts in the pre-render phase.
    """
    config = markata.get_plugin_config(__file__)

    if "description" not in config.keys():
        config["description"] = {}
        config["description"]["len"] = 160

    if "long_description" not in config.keys():
        config["long_description"] = {}
        config["long_description"]["len"] = 250

    def try_config_get(key: str) -> Any:
        try:
            return config.get(key).get("len") or None
        except AttributeError:
            return None

    max_description = max(
        [
            value
            for description_key in config
            if (value := try_config_get(description_key))
        ]
    )

    with markata.cache as cache:
        for article in markata.iter_articles("setting auto description"):
            set_description(
                markata=markata,
                article=article,
                cache=cache,
                config=config,
                max_description=max_description,
                plugin_text=Path(__file__).read_text(),
            )



"""manifest plugin"""
import json
from pathlib import Path
from typing import TYPE_CHECKING

from bs4 import BeautifulSoup

from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata.plugins.icon_resize import MarkataIcons


@hook_impl
def render(markata: "MarkataIcons") -> None:
    if "icons" in markata.__dict__.keys():
        icons = markata.icons
    else:
        icons = []
    manifest = {
        "name": markata.get_config("site_name") or "",
        "short_name": markata.get_config("short_name") or "",
        "start_url": markata.get_config("start_url") or "",
        "display": markata.get_config("display") or "",
        "background_color": markata.get_config("background_color") or "",
        "theme_color": markata.get_config("theme_color") or "",
        "description": markata.get_config("description") or "",
        "icons": icons,
    }
    filepath = Path(markata.config["output_dir"]) / "manifest.json"
    filepath.parent.mkdir(parents=True, exist_ok=True)
    filepath.touch(exist_ok=True)
    with open(filepath, "w+") as f:
        json.dump(manifest, f, ensure_ascii=True, indent=4)
    config = markata.get_plugin_config(__file__)
    with markata.cache as cache:
        for article in markata.iter_articles("add manifest link"):
            key = markata.make_hash(
                "seo",
                "manifest",
                article.content,
                article.html,
            )
            html_from_cache = cache.get(key)

            if html_from_cache is None:
                soup = BeautifulSoup(article.html, features="lxml")
                link = soup.new_tag("link")
                link.attrs["rel"] = "manifest"
                link.attrs["href"] = "/manifest.json"
                soup.head.append(link)

                html = soup.prettify()
                cache.add(key, html, expire=config["cache_expire"])
            else:
                html = html_from_cache
            article.html = html



"""Markata is a tool for handling directories of markdown.
"""
# annotations needed to return self
from __future__ import annotations

import datetime
import hashlib
import importlib
import os
import sys
import textwrap
from datetime import timedelta
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

import frontmatter
import pluggy
from checksumdir import dirhash
from diskcache import FanoutCache
from rich.console import Console
from rich.progress import track
from rich.table import Table

from markata import hookspec, standard_config
from markata.cli.plugins import Plugins
from markata.cli.runner import Runner
from markata.cli.server import Server
from markata.cli.summary import Summary
from markata.lifecycle import LifeCycle

__version__ = "0.2.0"


DEFAULT_MD_EXTENSIONS = [
    "codehilite",
    "markdown.extensions.admonition",
    "markdown.extensions.md_in_html",
    "markdown.extensions.tables",
    "markdown.extensions.toc",
    "pymdownx.betterem",
    "pymdownx.details",
    "pymdownx.emoji",
    "pymdownx.extra",
    "pymdownx.highlight",
    "pymdownx.inlinehilite",
    "pymdownx.keys",
    "pymdownx.magiclink",
    "pymdownx.saneheaders",
    "pymdownx.superfences",
    "pymdownx.tabbed",
    "pymdownx.tabbed",
    "pymdownx.tasklist",
    "pymdownx.tilde",
]

DEFAULT_HOOKS = [
    "markata.plugins.heading_link",
    "markata.plugins.pyinstrument",
    "markata.plugins.glob",
    "markata.plugins.load",
    "markata.plugins.auto_title",
    "markata.plugins.render_markdown",
    "markata.plugins.manifest",
    "markata.plugins.feeds",
    # "markata.plugins.generator",
    "markata.plugins.auto_description",
    "markata.plugins.seo",
    "markata.plugins.post_template",
    "markata.plugins.covers",
    "markata.plugins.copy_assets",
    "markata.plugins.publish_html",
    "markata.plugins.flat_slug",
    "markata.plugins.datetime",
    "markata.plugins.rss",
    "markata.plugins.icon_resize",
    "markata.plugins.sitemap",
    "markata.plugins.to_json",
    "markata.plugins.base_cli",
    "markata.plugins.tui",
    "markata.plugins.jinja_md",
]

DEFUALT_CONFIG = {
    "glob_patterns": ["**/*.md"],
    "hooks": ["default"],
    "disabled_hooks": [""],
    "markdown_extensions": [],
    "default_cache_expire": 3600,
    "output_dir": "markout",
    "assets_dir": "static",
}


class Post(frontmatter.Post):
    html: str


def set_phase(function: Callable) -> Any:
    def wrapper(self: Markata, *args: Tuple, **kwargs: Dict) -> Any:
        self.phase = function.__name__
        result = function(self, *args, **kwargs)
        self.phase = function.__name__
        self.phase_file.write_text(self.phase)
        return result

    return wrapper


class Markata:
    def __init__(self, console: Console = None) -> None:
        self.phase = "starting"
        self.MARKATA_CACHE_DIR = Path(".") / ".markata.cache"
        self.MARKATA_CACHE_DIR.mkdir(exist_ok=True)
        self.phase_file: Path = self.MARKATA_CACHE_DIR / "phase.txt"
        self.registered_attrs = hookspec.registered_attrs
        self.configure()
        if console is not None:
            self._console = console

    @property
    def cache(self) -> FanoutCache:
        return FanoutCache(self.MARKATA_CACHE_DIR, statistics=True)

    def __getattr__(self, item: str) -> Any:
        if item in self.__dict__.keys():
            return self.__getitem__(item)
        elif item in self.registered_attrs.keys():
            stage_to_run_to = max(
                [attr["lifecycle"] for attr in self.registered_attrs[item]]
            ).name
            self.run(stage_to_run_to)
            return getattr(self, item)
        else:
            raise AttributeError(item)

    @property
    def server(self) -> Server:
        try:
            return self._server
        except AttributeError:

            self._server: Server = Server(directory=str(self.config["output_dir"]))
            return self.server

    @property
    def runner(self) -> Runner:
        try:
            return self._runner
        except AttributeError:

            self._runner: Runner = Runner(self)
            return self.runner

    @property
    def plugins(self) -> Plugins:
        try:
            return self._plugins
        except AttributeError:

            self._plugins: Plugins = Plugins(self)
        return self.plugins

    @property
    def summary(self) -> Summary:
        try:
            return self._summary
        except AttributeError:

            self._summary: Summary = Summary(self)
            return self.summary

    def __rich__(self) -> Table:

        grid = Table.grid()
        grid.add_column("label")
        grid.add_column("value")

        for label, value in self.describe().items():
            grid.add_row(label, value)

        return grid

    def bust_cache(self) -> Markata:
        with self.cache as cache:
            cache.clear()
        return self

    @set_phase
    def configure(self) -> Markata:
        sys.path.append(os.getcwd())
        self.config = {**DEFUALT_CONFIG, **standard_config.load("markata")}
        if isinstance(self.config["glob_patterns"], str):
            self.config["glob_patterns"] = self.config["glob_patterns"].split(",")
        elif isinstance(self.config["glob_patterns"], list):
            self.config["glob_patterns"] = list(self.config["glob_patterns"])
        else:
            raise TypeError("glob_patterns must be list or str")
        self.glob_patterns = self.config["glob_patterns"]

        if "hooks" not in self.config:
            self.hooks = [""]
        if isinstance(self.config["hooks"], str):
            self.hooks = self.config["hooks"].split(",")
        if isinstance(self.config["hooks"], list):
            self.hooks = self.config["hooks"]

        if "disabled_hooks" not in self.config:
            self.disabled_hooks = [""]
        if isinstance(self.config["disabled_hooks"], str):
            self.disabled_hooks = self.config["disabled_hooks"].split(",")
        if isinstance(self.config["disabled_hooks"], list):
            self.disabled_hooks = self.config["disabled_hooks"]

        try:
            default_index = self.hooks.index("default")
            hooks = [
                *self.hooks[:default_index],
                *DEFAULT_HOOKS,
                *self.hooks[default_index + 1 :],
            ]
            self.hooks = [hook for hook in hooks if hook not in self.disabled_hooks]
        except ValueError:
            # 'default' is not in hooks , do not replace with default_hooks
            pass

        self._pm = pluggy.PluginManager("markata")
        self._pm.add_hookspecs(hookspec.MarkataSpecs)
        self._register_hooks()

        self._pm.hook.configure(markata=self)
        return self

    def get_plugin_config(self, path_or_name: str) -> Dict:

        key = Path(path_or_name).stem

        config = self.config.get(key, {})

        if not isinstance(config, dict):
            raise TypeError("must use dict")
        if "cache_expire" not in config.keys():
            config["cache_expire"] = self.config["default_cache_expire"]
        if "config_key" not in config.keys():
            config["config_key"] = key
        return config

    def get_config(
        self, key: str, warn: bool = True, suggested: Optional[str] = None
    ) -> Any:
        if key in self.config.keys():
            return self.config[key]
        else:

            if suggested is None:
                suggested = textwrap.dedent(
                    f"""
                        \[markata]
                        {key} = value
                    """  # noqa: W605
                )
            if warn:
                self.console.log(
                    textwrap.dedent(
                        f"""
                        Warning site_name is not set in markata config, sitemap will
                        be missing root site_name
                        to resolve this open your markata.toml and add

                        {suggested}

                        """
                    ),
                    style="yellow",
                )

    def make_hash(self, *keys: str) -> str:
        str_keys = [str(key) for key in keys]
        return hashlib.md5("".join(str_keys).encode("utf-8")).hexdigest()

    @property
    def phase(self) -> str:
        return self._phase

    @phase.setter
    def phase(self, value: str) -> None:
        self._phase = value

    @property
    def content_dir_hash(self) -> str:
        hashes = [dirhash(dir) for dir in self.content_directories]
        return self.make_hash(*hashes)

    @property
    def console(self) -> Console:
        try:
            return self._console
        except AttributeError:
            self._console = Console()
            return self._console

    def describe(self) -> dict[str, str]:
        return {"version": __version__, "phase": self.phase}

    def _to_dict(self) -> dict[str, Iterable]:
        return {"config": self.config, "articles": [a.to_dict() for a in self.articles]}

    def to_dict(self) -> dict:
        try:
            return self._to_dict()
        except AttributeError:
            self.render()
            return self._to_dict()

    def to_json(self) -> str:
        import json

        return json.dumps(self.to_dict(), indent=4, sort_keys=True, default=str)

    def _register_hooks(self) -> None:
        for hook in self.hooks:
            try:
                # module style plugins
                plugin = importlib.import_module(hook)
            except ModuleNotFoundError as e:
                # class style plugins
                if "." in hook:
                    mod = importlib.import_module(".".join(hook.split(".")[:-1]))
                    plugin = getattr(mod, hook.split(".")[-1])
                else:
                    raise e

            self._pm.register(plugin)

    def __iter__(self, description: str = "working...") -> Iterable[frontmatter.Post]:
        articles: Iterable[frontmatter.Post] = track(
            self.articles, description=description, transient=True, console=self.console
        )
        return articles

    def iter_articles(self, description: str) -> Iterable[frontmatter.Post]:
        articles: Iterable[frontmatter.Post] = track(
            self.articles, description=description, transient=True, console=self.console
        )
        return articles

    @set_phase
    def glob(self) -> Markata:
        """run glob hooks

        Glob hooks should append file lists to the markata object for later
        hooks to build from.  The default loader will utilize the `files`
        attribute for loading.
        """

        try:
            self._pm.hook.glob(markata=self)
        except AttributeError:
            self.configure()
            self._pm.hook.glob(markata=self)

        return self

    @set_phase
    def load(self) -> Markata:
        try:
            self._pm.hook.load(markata=self)
        except AttributeError:
            self.glob()
            self._pm.hook.load(markata=self)
        return self

    # @set_phase
    def pre_render(self) -> Markata:
        self._pm.hook.pre_render(markata=self)
        return self

    # @set_phase
    def render(self) -> Markata:
        try:
            self._pm.hook.pre_render(markata=self)
            self._pm.hook.render(markata=self)
            self._pm.hook.post_render(markata=self)
        except AttributeError:
            self.load()
            self._pm.hook.pre_render(markata=self)
            self._pm.hook.render(markata=self)
            self._pm.hook.post_render(markata=self)
        return self

    # @set_phase
    def post_render(self) -> Markata:
        self._pm.hook.post_render(markata=self)
        return self

    # @set_phase
    def save(self) -> Markata:
        try:
            self._pm.hook.save(markata=self)
        except AttributeError:
            self.render()
            self._pm.hook.save(markata=self)
        return self

    def run(self, lifecycle: LifeCycle = None) -> Markata:
        if lifecycle is None:
            lifecycle = getattr(LifeCycle, max(LifeCycle._member_map_))

        if isinstance(lifecycle, str):
            lifecycle = LifeCycle[lifecycle]

        stages_to_run = [m for m in LifeCycle._member_map_ if LifeCycle[m] <= lifecycle]

        self.console.log(f"running {stages_to_run}")
        for stage in stages_to_run:
            self.console.log(f"{stage} running")
            getattr(self, stage)()
            self.console.log(f"{stage} complete")

        with self.cache as cache:
            hits, misses = cache.stats()

        if hits + misses > 0:
            self.console.log(f"cache hit rate {round(hits/ (hits + misses)*100, 2)}%")
        self.console.log(f"cache hits/misses {hits}/{misses}")

        return self

    def filter(self, filter: str) -> List:
        def evalr(a: Post) -> Any:
            try:
                return eval(filter, {**a.to_dict(), "timedelta": timedelta}, {})
            except AttributeError:
                return eval(filter, {**a, "timedelta": timedelta}, {})

        return [a for a in self.articles if evalr(a)]

    def map(
        self,
        func: str = "title",
        filter: str = "True",
        sort: str = "True",
        reverse: bool = True,
        *args: Tuple,
        **kwargs: Dict,
    ) -> List:
        import copy

        def try_sort(a: Any) -> int:

            if "datetime" in sort.lower():
                return a.get(sort, datetime.datetime(1970, 1, 1))

            if "date" in sort.lower():
                return a.get(sort, datetime.date(1970, 1, 1))

            try:
                value = eval(sort, a.to_dict(), {})
            except NameError:
                return -1
            return value
            try:
                return int(value)
            except TypeError:
                try:
                    return int(value.timestamp())
                except Exception:
                    try:
                        return int(
                            datetime.datetime.combine(
                                value, datetime.datetime.min.time()
                            ).timestamp()
                        )
                    except Exception:
                        try:
                            return sum([ord(c) for c in str(value)])
                        except Exception:
                            return -1

        articles = copy.copy(self.articles)
        articles.sort(key=try_sort)
        if reverse:
            articles.reverse()

        return [
            eval(func, {**a.to_dict(), "timedelta": timedelta, "post": a}, {})
            for a in articles
            if eval(filter, {**a.to_dict(), "timedelta": timedelta, "post": a}, {})
        ]


def clif() -> None:
    import sys
    import time

    from rich import pretty, traceback

    if "--no-rich" not in sys.argv:
        pretty.install()
        traceback.install()

    m = Markata()

    if "--quiet" in sys.argv or "-q" in sys.argv:
        m.console.quiet = True
    else:
        m.console.print("console options:", m.console.options)

    if "--to-dict" in sys.argv:
        m.console.quiet = True
        data = m.to_dict()
        m.console.quiet = False
        m.console.print(data)
        return

    if "--draft" in sys.argv:
        print("\n".join([a["path"] for a in m.articles if a["status"] == "draft"]))

        return

    if "--today" in sys.argv:
        print("\n".join([a["path"] for a in m.articles if a["date"] == a["today"]]))
        return

    if "--scheduled" in sys.argv:
        print("\n".join([a["path"] for a in m.articles if a["date"] > a["today"]]))
        return

    if "--back-days" in sys.argv:
        print("\n".join([a["path"] for a in m.articles if a["date"] > a["today"]]))
        return

    if "--watch" in sys.argv:

        hash = m.content_dir_hash
        m.run()
        console = Console()
        with console.status("waiting for change", spinner="aesthetic", speed=0.2):
            while True:
                if m.content_dir_hash != hash:
                    hash = m.content_dir_hash
                    m.run()
                time.sleep(0.1)

    m.run()



"""Default glob plugin"""
from pathlib import Path
from typing import TYPE_CHECKING

from more_itertools import flatten

from markata.hookspec import hook_impl, register_attr

if TYPE_CHECKING:
    from markata import Markata


@hook_impl
@register_attr("content_directories", "files")
def glob(markata: "Markata") -> None:

    markata.files = list(
        flatten([Path().glob(str(pattern)) for pattern in markata.glob_patterns])
    )
    markata.content_directories = list(set([f.parent for f in markata.files]))

    try:
        ignore = markata.config["glob"]["use_gitignore"] or True
    except KeyError:
        ignore = True

    if ignore and (Path(".gitignore").exists() or Path(".markataignore").exists()):
        import pathspec

        lines = []

        if Path(".gitignore").exists():
            lines.extend(Path(".gitignore").read_text().splitlines())

        if Path(".markataignore").exists():
            lines.extend(Path(".markataignore").read_text().splitlines())

        spec = pathspec.PathSpec.from_lines("gitwildmatch", lines)

        markata.files = [
            file for file in markata.files if not spec.match_file(str(file))
        ]



"""Default load plugin."""
import time
from pathlib import Path
from typing import TYPE_CHECKING, Callable, List, Optional

import frontmatter
from rich.progress import BarColumn, Progress
from yaml.parser import ParserError

from markata.background import task
from markata.hookspec import hook_impl, register_attr

if TYPE_CHECKING:
    from markata import Markata, Post

    class MarkataMarkdown(Markata):
        articles: List = []


@hook_impl
@register_attr("articles")
def load(markata: "MarkataMarkdown") -> None:
    progress = Progress(
        BarColumn(bar_width=None), transient=True, console=markata.console
    )
    if not markata.config.get("repo_url", "https://github.com/").endswith("/"):
        markata.config["repo_url"] = (
            markata.config.get("repo_url", "https://github.com/") + "/"
        )

    futures = [get_post(article, markata) for article in markata.files]
    task_id = progress.add_task("loading markdown")
    progress.update(task_id, total=len(futures))
    with progress:
        while not all([f.done() for f in futures]):
            time.sleep(0.1)
            progress.update(task_id, total=len([f for f in futures if f.done()]))
    articles = [f.result() for f in futures]
    articles = [a for a in articles if a]
    markata.articles = articles


@task
def get_post(path: Path, markata: "Markata") -> Optional[Callable]:
    default = {
        "cover": "",
        "title": "",
        "tags": [],
        "status": "draft",
        "templateKey": "",
        "path": str(path),
        "description": "",
        "content": "",
    }
    try:
        post: "Post" = frontmatter.load(path)
        post.metadata = {**default, **post.metadata}
    except ParserError:
        return None
        post = default
    except ValueError:
        return None
        post = default
    post.metadata["path"] = str(path)
    post["edit_link"] = (
        str(markata.config.get("repo_url", "https://github.com/"))
        + "edit/"
        + str(markata.config.get("repo_branch", "main"))
        + "/"
        + str(post["path"])
    )
    return post



import time

import typer
from rich.layout import Layout


def make_layout() -> Layout:
    """Define the layout."""
    layout = Layout(name="root")

    layout.split(
        Layout(name="header", size=3),
        Layout(name="main"),
    )
    layout["main"].split_row(
        Layout(name="side", ratio=50),
        Layout(name="mid", ratio=30),
        Layout(name="describe", ratio=20),
    )
    layout["mid"].split(
        Layout(name="server"),
        Layout(name="runner"),
    )
    layout["side"].split(
        Layout(name="plugins"),
    )
    return layout


def run_until_keyboard_interrupt() -> None:
    try:
        while True:
            time.sleep(0.2)
    except KeyboardInterrupt:
        pass


def version_callback(value: bool) -> None:
    if value:
        from markata import __version__

        typer.echo(f"Markata CLI Version: {__version__}")
        raise typer.Exit()


def json_callback(value: bool) -> None:
    if value:
        from markata import Markata

        typer.echo(Markata().to_json())
        raise typer.Exit()


app = typer.Typer(
    name="Markata",
    help="Awesome CLI user manager.",
)


@app.callback()
def main(
    version: bool = typer.Option(
        None, "--version", callback=version_callback, is_eager=True
    ),
    to_json: bool = typer.Option(
        None, "--to-json", callback=json_callback, is_eager=True
    ),
) -> None:
    # Do other global stuff, handle other global options here
    return


def cli() -> None:
    from markata import Markata

    m = Markata()
    m._pm.hook.cli(markata=m, app=app)
    app()


if __name__ == "__main__":
    cli()



"""manifest plugin"""
from typing import TYPE_CHECKING, Dict, List

from bs4 import BeautifulSoup

from markata import Markata, __version__
from markata.hookspec import hook_impl

if TYPE_CHECKING:
    import frontmatter
    from bs4.element import Tag


def _create_seo(
    markata: Markata,
    soup: BeautifulSoup,
    article: "frontmatter.Post",
    site_name: str,
    author_name: str,
    author_email: str,
    twitter_card: str,
    twitter_creator: str,
    config_seo: Dict,
    images_url: str,
) -> List:
    if article.metadata["description"] == "" or None:
        try:
            article.metadata["description"] = " ".join(
                [p.text for p in soup.find(id="post-body").find_all("p")]
            ).strip()[:120]
        except AttributeError:
            article.metadata["description"] = ""

    seo = [
        *config_seo,
        {
            "name": "og:author",
            "property": "og:author",
            "content": author_name,
        },
        {
            "name": "og:author_email",
            "property": "og:author_email",
            "content": author_email,
        },
        {
            "name": "og:type",
            "property": "og:type",
            "content": "website",
        },
        {
            "name": "description",
            "property": "description",
            "content": article.metadata["description"],
        },
        {
            "name": "og:description",
            "property": "og:description",
            "content": article.metadata["description"],
        },
        {
            "name": "twitter:description",
            "property": "twitter:description",
            "content": article.metadata["description"],
        },
        {
            "name": "og:title",
            "property": "og:title",
            "content": f'{article.metadata["title"]} | {site_name}'[:60],
        },
        {
            "name": "twitter:title",
            "property": "twitter:title",
            "content": f'{article.metadata["title"]} | {site_name}'[:60],
        },
        {
            "name": "og:image",
            "property": "og:image",
            "content": f'{images_url}/{article.metadata["slug"]}-og.png',
        },
        {
            "name": "twitter:image",
            "property": "twitter:image",
            "content": f'{images_url}/{article.metadata["slug"]}-og.png',
        },
        {
            "name": "og:image:width",
            "property": "og:image:width",
            "content": "1600",
        },
        {
            "name": "og:image:width",
            "property": "og:image:width",
            "content": "900",
        },
        {
            "name": "twitter:card",
            "property": "twitter:card",
            "content": twitter_card,
        },
        {
            "name": "og:site_name",
            "property": "og:site_name",
            "content": site_name,
        },
        {
            "name": "twitter:creator",
            "property": "twitter:creator",
            "content": twitter_creator,
        },
        {
            "name": "title",
            "property": "title",
            "content": article.metadata["title"],
        },
        {
            "name": "generator",
            "property": "generator",
            "content": f"markata {__version__}",
        },
    ]
    return seo


def _add_seo_tags(seo: List, article: "frontmatter.Post", soup: BeautifulSoup) -> None:
    for meta in seo:
        soup.head.append(_create_seo_tag(meta, soup))


def _create_seo_tag(meta: dict, soup: BeautifulSoup) -> "Tag":
    tag = soup.new_tag("meta")
    for k in meta:
        tag.attrs[k] = meta[k]
    return tag


@hook_impl
def render(markata: Markata) -> None:

    url = markata.get_config("url") or ""
    images_url = markata.get_config("images_url") or url or ""
    site_name = markata.get_config("site_name") or ""
    author_name = markata.get_config("author_name") or ""
    author_email = markata.get_config("author_email") or ""
    twitter_creator = markata.get_config("twitter_creator") or ""
    twitter_card = markata.get_config("twitter_card") or "summary_large_image"
    config_seo = markata.get_config("seo", warn=False) or dict()

    with markata.cache as cache:
        for article in markata.iter_articles("add seo tags from seo.py"):
            key = markata.make_hash(
                "seo",
                "render",
                article.html,
                site_name,
                url,
                article.metadata["slug"],
                twitter_card,
                article.metadata["title"],
                str(config_seo),
            )

            html_from_cache = cache.get(key)

            if html_from_cache is None:
                soup = BeautifulSoup(article.html, features="lxml")
                seo = _create_seo(
                    markata=markata,
                    soup=soup,
                    article=article,
                    site_name=site_name,
                    author_name=author_name,
                    author_email=author_email,
                    twitter_card=twitter_card,
                    twitter_creator=twitter_creator,
                    config_seo=config_seo,
                    images_url=images_url,
                )
                _add_seo_tags(seo, article, soup)
                canonical_link = soup.new_tag("link")
                canonical_link.attrs["rel"] = "canonical"
                if article.metadata["slug"] == "index":
                    canonical_link.attrs["href"] = f"{url}/"
                else:
                    canonical_link.attrs["href"] = f'{url}/{article.metadata["slug"]}/'
                soup.head.append(canonical_link)

                meta_url = soup.new_tag("meta")
                meta_url.attrs["name"] = "og:url"
                meta_url.attrs["property"] = "og:url"
                if article.metadata["slug"] == "index":
                    meta_url.attrs["content"] = f"{url}/"
                else:
                    meta_url.attrs["content"] = f'{url}/{article.metadata["slug"]}/'
                soup.head.append(meta_url)

                html = soup.prettify()
                cache.add(key, html, expire=markata.config["default_cache_expire"])

            else:
                html = html_from_cache
            article.html = html



"""

### Configuration

Example configuration.  Covers supports multiple covers to be configured.  Here
is an example from my blog where we have a template sized for dev.to and one
sized for open graph.  Each image takes it's own configuration.

``` toml
[[markata.covers]]
name='-dev'
template = "static/cover-template.png"
font = "./static/JosefinSans-Regular.ttf"
text_font = "./static/JosefinSans-Regular.ttf"
font_color = "rgb(185,155,165)"
text_font_color = "rgb(255,255,255)"
text_key = 'description'
padding = [0, 40, 100, 300]
text_padding = [0,0]

[[markata.covers]]
name=''
template = "static/og-template.png"
font = "./static/JosefinSans-Regular.ttf"
font_color = "rgb(255,255,255)"
text_font = "./static/JosefinSans-Regular.ttf"
text_font_color = "rgb(200,200,200)"
text_key = 'description'
padding = [10, 10, 100, 300]
text_padding = [0,0]
```
"""
import time
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, List, Tuple, Union

from PIL import Image, ImageDraw, ImageFont
from rich.progress import BarColumn, Progress

from markata import background
from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata import Markata


@lru_cache(maxsize=64)
def _load_font(path: Path, size: int) -> ImageFont.FreeTypeFont:
    return ImageFont.truetype(path, size=size)


def get_font(
    path: Path,
    draw: ImageDraw.Draw,
    title: str,
    size: int = 250,
    max_size: tuple = (800, 220),
) -> ImageFont.FreeTypeFont:
    title = title or ""
    font = _load_font(path, size)
    current_size = draw.textsize(title, font=font)

    if current_size[0] > max_size[0] or current_size[1] > max_size[1]:
        return get_font(path, draw, title, size - 10, max_size=max_size)
    return font


class PaddingError(BaseException):
    def __init__(
        self,
        msg: str = "",
    ) -> None:
        super().__init__(
            "Padding must be an iterable of length 1, 2, 3, or 4.\n" + msg,
        )


def draw_text(
    image: Image,
    font_path: Path,
    text: str,
    color: Union[str, None],
    padding: Tuple[int, ...],
) -> None:
    text = text or ""
    draw = ImageDraw.Draw(image)
    padding = resolve_padding(padding)
    width = image.size[0]
    height = image.size[1]
    bounding_box = [padding[0], padding[1], width - padding[0], height - padding[1]]
    bounding_box = [padding[0], padding[1], width - padding[2], height - padding[3]]
    max_size = (bounding_box[2] - bounding_box[0], bounding_box[3] - bounding_box[1])
    x1, y1, x2, y2 = bounding_box
    font = get_font(font_path, draw, text, max_size=max_size)
    w, h = draw.textsize(text, font=font)
    x = (x2 - x1 - w) / 2 + x1
    y = (y2 - y1 - h) / 2 + y1
    draw.text((x, y), text, fill=color, font=font, align="center")


def resolve_padding(padding: Tuple[int, ...]) -> Tuple[int, ...]:
    """Convert padding to a len 4 tuple"""
    if len(padding) == 4:
        return padding
    if len(padding) == 3:
        return (*padding, padding[1])
    if len(padding) == 2:
        return padding * 2
    if len(padding) == 1:
        return padding * 4
    raise PaddingError(f"recieved padding: {padding}")


@background.task
def make_cover(
    title: str,
    color: str,
    output_path: Path,
    template_path: Path,
    font_path: Path,
    padding: Tuple[int, ...],
    text_font: Path,
    text: str = None,
    text_font_color: str = None,
    text_padding: Tuple[int, ...] = None,
    resizes: List[int] = None,
) -> None:
    if output_path.exists():
        return
    image = Image.open(template_path)
    draw_text(image, font_path, title, color, padding)
    if text is not None:
        if text_padding is None:
            text_padding = (
                image.size[1] - image.size[1] / 5,
                image.size[0] / 5,
                image.size[1] - image.size[1] / 10,
            )
        draw_text(image, text_font, text, text_font_color, text_padding)

    output_path.parent.mkdir(exist_ok=True)
    image.save(output_path)
    ratio = image.size[1] / image.size[0]

    covers = []
    if resizes:
        for width in resizes:

            re_img = image.resize((width, int(width * ratio)), Image.ANTIALIAS)
            filename = (
                f"{output_path.stem}_{width}x{int(width*ratio)}{output_path.suffix}"
            )
            covers.append(filename)

            filepath = Path(output_path.parent / filename)
            re_img.save(filepath)


@hook_impl
def save(markata: "Markata") -> None:
    futures = []

    if "covers" not in markata.config.keys():
        return

    for article in markata.iter_articles("making covers"):
        for cover in markata.config["covers"]:
            try:
                padding = cover["padding"]
            except KeyError:
                padding = (
                    200,
                    100,
                )
            try:
                text_padding = cover["text_padding"]
            except KeyError:
                text_padding = (
                    200,
                    100,
                )
            if "text_key" in cover:
                try:
                    text = article.metadata[cover["text_key"]]
                except AttributeError:
                    text = article[cover["text_key"]]
                try:
                    text = text.replace("\n", "")
                    from more_itertools import chunked

                    text = "\n".join(["".join(c) for c in chunked(text, 60)])
                except AttributeError:
                    # text is likely None
                    pass

                text_font = cover["text_font"]
                text_font_color = cover["text_font_color"]
            else:
                text = None
                text_font = None
                text_font_color = None
            try:
                title = article.metadata["title"]
            except AttributeError:
                title = article["title"]
            futures.append(
                make_cover(
                    title=title,
                    color=cover["font_color"],
                    output_path=Path(markata.config["output_dir"])
                    / (article["slug"] + cover["name"] + ".png"),
                    template_path=cover["template"],
                    font_path=cover["font"],
                    padding=padding,
                    text_font=text_font,
                    text=text,
                    text_font_color=text_font_color,
                    text_padding=text_padding,
                    resizes=cover.get("resizes"),
                )
            )

    progress = Progress(
        BarColumn(bar_width=None), transient=True, console=markata.console
    )
    task_id = progress.add_task("loading markdown")
    progress.update(task_id, total=len(futures))
    with progress:
        while not all([f.done() for f in futures]):
            time.sleep(0.1)
            progress.update(task_id, total=len([f for f in futures if f.done()]))
    [f.result() for f in futures]



from datetime import datetime

from rich.panel import Panel
from rich.table import Table


class Header:
    """Display header with clock."""

    def __rich__(self) -> Panel:
        grid = Table.grid(expand=True)
        grid.add_column(justify="center", ratio=1)
        grid.add_column(justify="right")
        grid.add_row(
            "[magenta][b]Markata[/b][/] [bright_black]Live Server[/]",
            datetime.now().ctime(),
        )
        return Panel(grid, style="yellow")



import time
from typing import TYPE_CHECKING

from rich.panel import Panel
from rich.text import Text

if TYPE_CHECKING:
    from markata import Markata


class Runner:
    """Display Footer"""

    _status = "waiting"

    _dirhash = ""
    time = time.time()

    def __init__(self, markata: "Markata") -> None:
        self.m = markata

    def run(self) -> None:
        self.status = "running"
        self.m.run()
        self.time = time.time()
        self.status = "waiting"

    def __rich__(self) -> Panel:

        if self._dirhash != self.m.content_dir_hash:
            self.run()
            self._dirhash = self.m.content_dir_hash

        s = f"runner is waiting {round(time.time() - self.time)}"

        return Panel(Text(s), border_style="green", title="runner")


if __name__ == "__main__":
    from rich.live import Live

    from .cli import run_until_keyboard_interrupt

    with Live(Runner(Markata()), refresh_per_second=30, screen=True):
        run_until_keyboard_interrupt()



"""Icon Resize Plugin"""
from pathlib import Path
from typing import TYPE_CHECKING

from PIL import Image

from markata.hookspec import hook_impl, register_attr

if TYPE_CHECKING:

    from typing import Dict, List

    from markata import Markata

    class MarkataIcons(Markata):
        icons: List[Dict[str, str]]


@hook_impl
@register_attr("icons")
def render(markata: "MarkataIcons") -> None:
    if "icon" not in markata.config:
        return
    base_out_file = Path(markata.config["output_dir"]) / markata.config["icon"]

    with Image.open(Path(markata.config["assets_dir"]) / markata.config["icon"]) as img:
        markata.icons = []
        for width in [48, 72, 96, 144, 192, 256, 384, 512]:
            height = int(float(img.size[1]) * float(width / float(img.size[0])))
            filename = Path(
                f"{base_out_file.stem}_{width}x{height}{base_out_file.suffix}"
            )
            markata.icons.append(
                {
                    "src": str(filename),
                    "sizes": f"{width}x{width}",
                    "type": f"image/{img.format}".lower(),
                    "purpose": "any maskable",
                }
            )


@hook_impl
def save(markata: "MarkataIcons") -> None:
    if "icon" not in markata.config:
        return
    base_out_file = Path(markata.config["output_dir"]) / markata.config["icon"]
    for width in [48, 72, 96, 144, 192, 256, 384, 512]:
        with Image.open(
            Path(markata.config["assets_dir"]) / markata.config["icon"]
        ) as img:
            height = int(float(img.size[1]) * float(width / float(img.size[0])))
            img = img.resize((width, height), Image.ANTIALIAS)
            filename = Path(
                f"{base_out_file.stem}_{width}x{height}{base_out_file.suffix}"
            )
            out_file = Path(markata.config["output_dir"]) / filename
            img.save(out_file)



from typing import TYPE_CHECKING, List

import markdown

from markata import DEFAULT_MD_EXTENSIONS
from markata.hookspec import hook_impl, register_attr

if TYPE_CHECKING:
    from markata import Markata

    class MarkataMarkdown(Markata):
        articles: List = []
        md: markdown.Markdown = markdown.Markdown()
        markdown_extensions: List = []


@hook_impl(tryfirst=True)
@register_attr("md", "markdown_extensions")
def configure(markata: "MarkataMarkdown") -> None:
    if "markdown_extensions" not in markata.config:
        markdown_extensions = [""]
    if isinstance(markata.config["markdown_extensions"], str):
        markdown_extensions = [markata.config["markdown_extensions"]]
    if isinstance(markata.config["markdown_extensions"], list):
        markdown_extensions = markata.config["markdown_extensions"]
    else:
        raise TypeError("markdown_extensions should be List[str]")

    markata.markdown_extensions = [*DEFAULT_MD_EXTENSIONS, *markdown_extensions]
    markata.md = markdown.Markdown(extensions=markata.markdown_extensions)


@hook_impl(tryfirst=True)
def render(markata: "Markata") -> None:
    config = markata.get_plugin_config(__file__)
    with markata.cache as cache:
        for article in markata.iter_articles("rendering markdown"):
            key = markata.make_hash(
                "render_markdown",
                "render",
                article.content,
            )
            html_from_cache = cache.get(key)
            if html_from_cache is None:
                html = markata.md.convert(article.content)
                cache.add(key, html, expire=config["cache_expire"])
            else:
                html = html_from_cache
            article.html = html
            article.article_html = html



from pathlib import Path
from typing import TYPE_CHECKING

import frontmatter

from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata import Markata

DEV_TO_FRONTMATTER = [
    "title",
    "published",
    "description",
    "tags",
    "canonical_url",
    "cover_image",
    "series",
]


def should_join(line):
    if line == "":
        return False
    if line is None:
        return False
    if line[0].isalpha():
        return True
    if line[0].startswith("["):
        return True
    if line[0].startswith("!"):
        return True
    return False


def join_lines(article):
    lines = article.split("\n")
    line_number = 0
    while line_number + 1 < len(lines):
        line = lines[line_number]
        nextline = lines[line_number + 1]
        if should_join(line) and should_join(nextline):
            lines[line_number] = f"{line} {nextline}"
            lines.pop(line_number + 1)
        else:
            line_number += 1

    return "\n".join(lines)


@hook_impl
def post_render(markata: "Markata") -> None:

    for post in markata.iter_articles(description="saving source documents"):

        article = frontmatter.Post(
            post.content,
            **{k: v for k, v in post.metadata.items() if k in DEV_TO_FRONTMATTER},
        )

        article.content = join_lines(article.content)
        article.content = join_lines(article.content)

        if "canonical_url" not in article:
            article["canonical_url"] = f'{markata.config["url"]}/{post["slug"]}/'

        if "published" not in article:
            article["published"] = True

        if "cover_image" not in article:
            article[
                "cover_image"
            ] = f"{markata.config['images_url']}/{post['slug']}.png"
        post.dev_to = article


@hook_impl
def save(markata: "Markata") -> None:
    output_dir = Path(str(markata.config["output_dir"]))
    output_dir.mkdir(parents=True, exist_ok=True)
    for post in markata.iter_articles(description="saving source documents"):

        with open(output_dir / Path(post["slug"]) / "dev.md", "w+") as f:
            f.write(frontmatter.dumps(post.dev_to))



"""Default glob plugin"""
from pathlib import Path
from typing import TYPE_CHECKING

from feedgen.feed import FeedGenerator

from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata import Markata

    class MarkataRss(Markata):
        fg: "FeedGenerator"
        rss: str


@hook_impl(trylast=True)
def render(markata: "MarkataRss") -> None:
    fg = FeedGenerator()
    url = markata.get_config("url") or ""
    title = markata.get_config("title") or "rss_feed"
    name = markata.get_config("author_name") or ""
    email = markata.get_config("author_email") or ""
    icon = markata.get_config("icon") or ""
    lang = markata.get_config("lang") or ""
    rss_description = markata.get_config("rss_description") or "rss_feed"

    fg.id(url + "/rss.xml")
    fg.title(title)
    fg.author(
        {
            "name": name,
            "email": email,
        }
    )
    fg.link(href=url, rel="alternate")
    fg.logo(icon)
    fg.subtitle(rss_description)
    fg.link(href=url + "/rss.xml", rel="self")
    fg.language(lang)

    try:
        all_posts = reversed(sorted(markata.articles, key=lambda x: x["date"]))
        posts = [post for post in all_posts if post["status"] == "published"]
    except BaseException:
        posts = markata.articles

    for article in posts:
        fe = fg.add_entry()
        fe.id(url + "/" + article["slug"])
        fe.title(article.metadata["title"])
        fe.published(article.metadata["datetime"])
        fe.description(article.metadata["description"])
        fe.summary(article.metadata["long_description"])
        fe.link(href=url + "/" + article["slug"])
        fe.content(article.article_html.translate(dict.fromkeys(range(32))))

    markata.fg = fg
    markata.rss = fg.rss_str(pretty=True)


@hook_impl
def save(markata: "MarkataRss") -> None:
    output_dir = Path(markata.config["output_dir"])
    markata.fg.rss_file(str(output_dir / "rss.xml"))



"""Define hook specs."""
import functools
from typing import TYPE_CHECKING, Any, Callable, Dict, List

import pluggy

from markata.lifecycle import LifeCycle

if TYPE_CHECKING:

    import typer

    from markata import Markata

HOOK_NAMESPACE = "markata"
hook_spec = pluggy.HookspecMarker(HOOK_NAMESPACE)
hook_impl = pluggy.HookimplMarker(HOOK_NAMESPACE)


class MarkataSpecs:
    """
    Namespace that defines all specifications for Load hooks.

    configure -> glob -> load -> render -> save
    """


@hook_spec
def generic_lifecycle_method(
    markata: "Markata",
) -> Any:
    ...


@hook_spec
def cli_lifecycle_method(markata: "Markata", app: "typer.Typer") -> Any:
    "A Markata lifecycle methos that includes a typer app used for cli's"


for method in LifeCycle._member_map_:
    if "cli" in method:
        setattr(MarkataSpecs, method, cli_lifecycle_method)
    else:
        setattr(MarkataSpecs, method, generic_lifecycle_method)

registered_attrs: Dict[str, List[Dict[str, Any]]] = {}


def register_attr(*attrs: Any) -> Callable:
    def decorator_register(
        func: Callable,
    ) -> Callable:

        for attr in attrs:
            if attr not in registered_attrs:
                registered_attrs[attr] = []
            registered_attrs[attr].append(
                {
                    "func": func,
                    "funcname": func.__code__.co_name,
                    "lifecycle": getattr(LifeCycle, func.__code__.co_name),
                }
            )

        @functools.wraps(func)
        def wrapper_register(markata: "Markata", *args: Any, **kwargs: Any) -> Any:
            return func(markata, *args, **kwargs)

        return wrapper_register

    return decorator_register



"""
leading docstring
"""
import ast
import datetime
import textwrap
from pathlib import Path
from typing import TYPE_CHECKING, List

import frontmatter

from markata.hookspec import hook_impl, register_attr

if TYPE_CHECKING:
    from markata import Markata

    class MarkataDocs(Markata):
        py_files: List = []
        content_directories: List = []


def add_parents(tree: ast.AST) -> None:
    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            child.parent = node
            if not hasattr(child, "parents"):
                child.parents = [node]
            child.parents.append(node)
            if isinstance(node, ast.ClassDef) and isinstance(child, ast.FunctionDef):
                child.type = "method"
            elif isinstance(child, ast.FunctionDef):
                child.type = "function"
            elif isinstance(child, ast.ClassDef):
                child.type = "class"


@hook_impl
@register_attr("content_directories", "py_files")
def glob(markata: "MarkataDocs") -> None:
    """
    finds k

    ## Parameters

    `markata` the markata object

    """

    markata.py_files = list(Path().glob("**/*.py"))

    content_directories = list(set([f.parent for f in markata.py_files]))
    if "content_directories" in markata.__dict__.keys():
        markata.content_directories.extend(content_directories)
    else:
        markata.content_directories = content_directories

    try:
        ignore = markata.config["glob"]["use_gitignore"] or True
    except KeyError:
        ignore = True

    if ignore and (Path(".gitignore").exists() or Path(".markataignore").exists()):
        import pathspec

        lines = []

        if Path(".gitignore").exists():
            lines.extend(Path(".gitignore").read_text().splitlines())

        if Path(".markataignore").exists():
            lines.extend(Path(".markataignore").read_text().splitlines())

        spec = pathspec.PathSpec.from_lines("gitwildmatch", lines)

        markata.py_files = [
            file for file in markata.py_files if not spec.match_file(str(file))
        ]


def make_article(markata: "Markata", file: Path) -> frontmatter.Post:
    raw_source = file.read_text()
    tree = ast.parse(raw_source)
    add_parents(tree)
    nodes = [
        n
        for n in ast.walk(tree)
        if isinstance(n, ast.FunctionDef) or isinstance(n, ast.ClassDef)
    ]

    edit_link = (
        str(markata.config.get("repo_url", "https://github.com/"))
        + "edit/"
        + str(markata.config.get("repo_branch", "main"))
        + "/"
        + str(file)
    )
    article = textwrap.dedent(
        f"""
    ---
    title: {file.name}
    status: published
    slug: {file.parent}/{file.stem}
    edit_link: {edit_link}
    path: {file.stem}.md
    today: {datetime.datetime.today()}
    description: Docs for {file.stem}

    ---

    """
    )
    article += textwrap.dedent(
        f"""
            {ast.get_docstring(tree) or ""}
            """
    )
    for node in nodes:
        article += textwrap.dedent(
            f"""

---

## {node.name} `{node.type}`

{ast.get_docstring(node)}

??? "{node.name} source"
    ``` python
{textwrap.indent(ast.get_source_segment(raw_source, node), '    ')}
    ```
        """
        )

    return frontmatter.loads(article)


@hook_impl
@register_attr("articles")
def load(markata: "MarkataDocs") -> None:
    """
    similar to [glob](../glob)
    """
    if "articles" not in markata.__dict__:
        markata.articles = []
    for py_file in markata.py_files:
        markata.articles.append(make_article(markata, py_file))



"""
Markata plugin to create a pyinstrument profile if pyinstrument is installed.

The profile will be saved to <output_dir>/_profile/index.html
"""
from pathlib import Path

from markata import Markata
from markata.hookspec import hook_impl, register_attr

try:
    from pyinstrument import Profiler
except ModuleNotFoundError:
    "ignore if pyinstrument does not exist"
    ...


class MarkataInstrument(Markata):
    should_profile = False
    profiler = None


@hook_impl(tryfirst=True)
@register_attr("should_profile")
def configure(markata: MarkataInstrument) -> None:
    "set the should_profile variable"

    if "should_profile" not in markata.__dict__.keys():
        try:
            markata.should_profile = markata.config["pyinstrument"]["should_profile"]
        except KeyError:
            markata.should_profile = False


@hook_impl(tryfirst=True)
@register_attr("profiler")
def glob(markata: MarkataInstrument) -> None:
    "start the profiler as soon as possible"
    if markata.should_profile:
        try:
            markata.profiler = Profiler()
            markata.profiler.start()
        except NameError:
            "ignore if Profiler does not exist"
            ...


@hook_impl(trylast=True)
def save(markata: MarkataInstrument) -> None:
    "stop the profiler and save as late as possible"
    if markata.should_profile:
        try:

            if "profiler" in markata.__dict__.keys():
                output_file = (
                    Path(markata.config["output_dir"]) / "_profile" / "index.html"
                )
                output_file.parent.mkdir(parents=True, exist_ok=True)
                markata.profiler.stop()
                html = markata.profiler.output_html()
                output_file.write_text(html)
                markata.console.print(markata.profiler.output_text())

        except AttributeError:
            "ignore if markata does not have a profiler attribute"
            ...



from pathlib import Path
from typing import TYPE_CHECKING

from PIL import Image, ImageDraw, ImageFont

from markata import background
from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata import Markata


def get_font(
    path: Path, draw: ImageDraw.Draw, title: str, size: int = 250
) -> ImageFont.FreeTypeFont:
    font = ImageFont.truetype(path, size=size)
    if draw.textsize(title, font=font)[0] > 800:
        return get_font(path, draw, title, size - 10)
    return font


@background.task
def make_cover(
    title: str, color: str, output_path: Path, template_path: Path, font_path: Path
) -> None:
    image = Image.open(template_path)

    draw = ImageDraw.Draw(image)

    font = get_font(font_path, draw, title)

    color = "rgb(255,255,255)"
    padding = (200, 100)
    bounding_box = [padding[0], padding[1], 1000 - padding[0], 420 - padding[1]]
    x1, y1, x2, y2 = bounding_box
    w, h = draw.textsize(title, font=font)
    x = (x2 - x1 - w) / 2 + x1
    y = (y2 - y1 - h) / 2 + y1
    draw.text((x, y), title, fill=color, font=font, align="center")
    image.save(output_path)


@hook_impl
def save(markata: "Markata") -> None:
    for article in markata.articles:
        output_path = Path(markata.output_dir) / (
            Path(article.metadata["path"]).stem + ".png"
        )

        make_cover(
            article.metadata["title"],
            markata.config["cover_font_color"],
            output_path,
            markata.config["cover_template"],
            markata.config["cover_font"],
        )



"""
Markata's base command line commands.

This plugin enables
[`build`](https://markata.dev/markata/plugins/base_cli/#build-function)
and
[`list`](https://markata.dev/markata/plugins/base_cli/#list-function)
commands as part of the main markata cli.

## Building Your Site with the Cli

Your Markata Site can be build completely from the command line.

``` bash
markata build

# or if you prefer pipx
pipx run markata build
```

see the
[`build`](https://markata.dev/markata/plugins/base_cli/#build-function)
section for more examples.

## Listing your articles

Markata list is a tool to help list out artile attributes right to your
terminal.  This is very helpful to find articles on larger sites, or
debug what is getting picked up by markata.

``` bash
markata list --map 'str(date.year) + "," + title'
```

see the
[`list`](https://markata.dev/markata/plugins/base_cli/#list-function)
section for more examples.
"""
import pdb
import shutil
import sys
import traceback
import warnings
from typing import TYPE_CHECKING, Callable, Optional

import typer

from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata import Markata


def make_pretty() -> None:
    """
    This is a helper function that enables suppresses tracebacks from
    frameworks like `click` that can make your traceback long and hard
    to follow.  It also makes evrerything more colorful and easier to
    follow.
    """
    import click
    import pluggy
    import typer
    from rich import pretty as _pretty
    from rich import traceback

    _pretty.install()
    traceback.install(
        show_locals=True,
        suppress=[
            pluggy,
            click,
            typer,
        ],
    )


@hook_impl()
def cli(app: typer.Typer, markata: "Markata") -> None:
    """
    Markata hook to implement base cli commands.
    """

    @app.command()
    def build(
        pretty: bool = True,
        quiet: bool = typer.Option(
            False,
            "--quiet",
            "-q",
        ),
        # to_dict: bool = False,
        verbose: bool = typer.Option(
            False,
            "--verbose",
            "-v",
        ),
        should_pdb: bool = typer.Option(
            False,
            "--pdb",
        ),
        profile: bool = True,
    ) -> None:
        """
        Markata's primary way of building your site for production.
        By default, running `markta build` will render your markdown to
        the `./markout` directory.

        ``` bash
        markata build
        ```

        If you are having an issue and want to pop immediately into a debugger
        upon failure you can pass the `--pdb` flag to the build command.

        ``` bash
        markata build  --pdb
        ```

        If you do not like the way rich looks, or its suppressing tracebaks you
        would like to remain visible you can use `--no-pretty`

        ``` bash
        markata build --no-pretty
        ```

        If you need to run without any console logging pass in the
        `--quiet` flag.

        ``` bash
        markata build --quiet
        ```

        `markta build` will automatically run the pyinstrument profiler
        while building your site if you have pyinstrument installed.  It
        will echo out your profile in the console as well as write it to
        `/_profile` on your built site. If you prefer not to run
        pyinstrument profiling, even when it is installed you can pass
        in `--no-profile`

        ``` bash
        markata build --no-profile
        ```
        """

        if pretty:
            make_pretty()

        if quiet:
            markata.console.quiet = True

        if verbose:
            markata.console.print("console options:", markata.console.options)

        if profile:
            markata.should_profile_cli = True
            markata.should_profile = True
            markata.configure()

        if should_pdb:
            pdb_run(markata.run)

        else:
            markata.run()

    @app.command()
    def list(
        map: str = "title",
        filter: str = "True",
        sort: str = "True",
        head: Optional[int] = None,
        tail: Optional[int] = None,
        include_empty: bool = False,
        reverse: bool = False,
        use_pager: bool = typer.Option(True, "--pager", "--no-pager"),
    ) -> None:
        """
        Provides a way run markatas, map, filter, and sort from the
        command line.  I personally use this more often than the build
        command while I am writing on a site with a large number of
        posts on it.  It makes slicing in by `templatekey`, `tag`, or
        `date` much easier.

        ### default list

        By default `markata list` will list all titles in a pager, for all posts
        being loaded by markata.

        ``` bash
        markata list
        ```

        ### Skip the pager

        Markata uses rich for its pager, it's pretty smart about when to
        use the pager or pass text to the next thing in the pipeline,
        but if you don't want to run a pager you can pass  `--no-pager`

        ``` bash
        markata list --no-pager
        ```

        ### List other attributes

        You can list any other attribute tied to your posts.  These are
        added through either your yaml frontmatter at the start of your
        post, or through the use of a plugin.


        ``` bash
        # the filepath of the post
        markata list --map path

        # the slug of the post (where it will show up on the site)
        markata list --map slug

        # the date of the post
        markata list --map date

        # the full raw content of the post
        markata list --map content
        ```

        ### List more than one attribute

        You can create new attributes as you map to echo out by
        combining existing attributes.

        ``` bash
        markata list --map 'title + " , " + slug'
        ```

        ### Using Python objects as map

        You can access attributes of each post attribute that you map
        over.  For instance on my blog, each post has a date that is a
        datetime object.  I can ask each post for its `date.year`

        ``` bash
        markata list --map date.year

        # combining this with title
        markata list --map 'str(date.year) + "," + title'
        ```

        ### Filtering posts

        Posts are filtered with python syntax, you will have all
        attributes tied to your posts available to filter with.

        ``` bash
        markata list --filter "'__' not in title"
        ```

        ### Filtering by dates

        If your site has dates tied to your posts you can filter by
        date.  On my blog this makes a ton of sense and is quite useful.
        On the Markata docs though it doesn't really make much sense,
        since there really isn't the idea of a post date there.

        ``` bash
        # listing today's posts
        markata list --filter "date==today"

        # listing this year's posts
        markata list --filter "date.year==today.year"
        ```

        ### Full Content Search

        You can also search the full content of each post for specific
        words.
        ``` bash

        markata list --filter "'python' in content"
        ```

        ### Filtering by frontmatter data

        I use a templateKey on my personal blog to determine which
        template to render the page with.  I can fitler my posts by a
        `til` (today i learned) key.

        ``` bash
        markata list --filter "templateKey=='til'"
        ```

        ### Combining filters

        Filters can be combined together quite like maps can, it's all
        just python syntax.

        ``` bash
        markata list --filter "templateKey=='til' and date == today"
        ```

        ### Sorting posts

        Posts can be sorted by attributes on your post, and they can
        even be reversed.

        ``` bash
        markta list --sort date
        markta list --sort date --reverse
        ```

        ### Putting it all together

        The real power of all this comes when you combine them all into
        lists that work for you and your workflow.  This really makes
        working on larger projects so much easier to find things.


        ### Making a fuzzy picker for your posts

        Here is a bash command to open an fzf picker for todays posts,
        then open it in your `$EDITOR`

        ``` bash
        markata list --map path --filter 'date==today' --sort date --reverse | fzf --preview 'bat --color always {}' | xargs -I {} $EDITOR {}
        ```

        ### Combining wtih nvim Telescope

        Here is the same command setup as a Telescope picker for neovim.

        ``` vim
        nnoremap <leader>et <cmd>Telescope find_files find_command=markata,list,--map,path,--filter,date==today<cr>
        ```

        If you have another way to open posts in your editor with
        `markata list` I would love to accept a PR to add it to the
        examples here.
        """

        markata.console.quiet = True

        tail = -tail if tail else tail
        filtered = markata.map(map, filter, sort)
        if not include_empty:
            filtered = [a for a in filtered if a != ""]
        filtered = filtered[tail:head]
        if reverse:
            filtered = reversed(filtered)

        markata.console.quiet = False
        if markata.console.is_terminal and use_pager:
            with markata.console.pager():
                for a in filtered:
                    markata.console.print(a, style="purple")
        else:
            for a in filtered:
                markata.console.print(a)

    @app.command()
    def clean(
        quiet: bool = typer.Option(
            False,
            "--quiet",
            "-q",
        ),
        dry_run: bool = typer.Option(
            False,
            "--dry-run",
        ),
    ):
        """
        Cleans up output generated by markata including both the output_dir and the .markata_cache.

        ### Dry Run

        You can run with `--dry-run` to see what markata is about to do.

        ``` bash
        markata clean --dry-run
        [09:42:37] [DRYRUN] removing outptut directory: markout                                                                                                      base_cli.py:371
                   [DRYRUN] removing cache directory: .markata.cache                                                                                                 base_cli.py:377

        ```

        ### Running clean

        Running markata clean will fully delete all of the directories created
        by markata.

        ``` bash
        markata clean
        [09:53:04]  removing outptut directory: markout                                                                                                              base_cli.py:394
                    removing cache directory: .markata.cache                                                                                                         base_cli.py:405
        ```

        ### Running Quietly

        Running with `--quiet` will remove all of the directories created by
        markata without announcing what it is doing.

        ``` bash
        markata clean --quiet
        ```
        """
        _clean(markata=markata, quiet=quiet, dry_run=dry_run)


def _clean(markata, quiet: bool = False, dry_run: bool = False):
    if quiet:
        markata.console.quiet = True

    markata.console.log(
        f'{"[DRYRUN]" if dry_run else ""} removing outptut directory: {markata.config.get("output_dir")}'
    )
    if not dry_run:
        try:
            shutil.rmtree(str(markata.config.get("output_dir")))
        except FileNotFoundError:
            warnings.warn(
                f'output directory: {markata.config.get("output_dir")} does not exist'
            )

    markata.console.log(
        f'{"[DRYRUN]" if dry_run else ""} removing cache directory: .markata.cache'
    )
    if not dry_run:
        try:
            shutil.rmtree(".markata.cache")
        except FileNotFoundError:
            warnings.warn("cache directory: .markata.cache does not exist")


def pdb_run(func: Callable) -> None:
    """
    Wraps a function call with a post_mortem pdb debugger.
    """
    try:
        func()
    except Exception:
        extype, value, tb = sys.exc_info()
        traceback.print_exc()
        pdb.post_mortem(tb)



"""Default datetime plugin"""
import datetime
from typing import TYPE_CHECKING

import dateutil.parser
import pytz

from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata import Markata


@hook_impl(trylast=True)
def load(markata: "Markata") -> None:
    for article in markata.iter_articles("datetime"):

        try:
            date = article.metadata["date"]
        except KeyError:
            date = None
        if isinstance(date, str):
            date = dateutil.parser.parse(date)
        if isinstance(date, datetime.date):
            date = datetime.datetime(
                year=date.year,
                month=date.month,
                day=date.day,
                tzinfo=pytz.utc,
            )

        article["today"] = datetime.date.today()
        article["now"] = datetime.datetime.now()
        article["datetime"] = date
        if date is not None:
            article["date"] = date.date()



"""
The articles slug is what determines the url to your page.  It should be
sanitized of special characters that do not work in the browser.


## Explicit Slug in Frontmatter

If you explicitly set the slug in the frontmatter of a post, markata will not
overwrite it.

```markdown
---
title: My First Post
slug: /my-post

---

This is my first post it will be at `<markata.config['url']>/my-post/`
reguardless of filename.
```

## Automatic Slug Based on Filename

By default the flat_slug plugin will use the `stem` of your filename, which is
the filename without the extension, unless you explicitly set your slug in
frontmatter.

* `/pages/my-post.md` becomes `<markata.config['url']>/my-post/`
* `/pages/blog/a-blog-post.md` becomes `<markata.config['url']>/a-blog-post/`
"""
from pathlib import Path
from typing import TYPE_CHECKING

from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata import Markata


@hook_impl(tryfirst=True)
def pre_render(markata: "Markata") -> None:
    """
    Sets the article slug if one is not already set in the frontmatter.
    """
    for article in markata.iter_articles(description="creating slugs"):
        article["slug"] = article.get("slug", Path(article["path"]).stem)



"""
The prevnext plugin, creates previous and next links inside each post.

## Example config

In this example we have two maps of posts to look through.  prevnext will look
through each of these lists of posts for the current post, then return the post
before and after this post as the prevnext posts.

``` toml

[markata]
# default colors will be taken from markata's color_text and color_accent
color_text=white
color_text_light=black
color_accent=white
color_accent_light=black

[markata.prevnext]
# strategy can be 'first' or 'all'
# 'first' will cycle through the first map the post is found in.
# 'all' will cycle through all of the maps
strategy='first'

# if you want different colors than your main color_text and color_accent, then
# you can override it here
# colors can be any valid css color format

prevnext_color_text=white
prevnext_color_text_light=black
prevnext_color_angle=white
prevnext_color_angle_light=black


# you can have multiple maps, the order they appear will determine their preference
[markata.feeds.python]
filter='"python" in tags'
sort='slug'

[markata.feeds.others]
filter='"python" not in tags'
sort='slug'
```

The configuration below will setup two maps, one where "python" is in the list
of tags, and another where it is not.  This will link all python posts together
with a prevnext cycle, and all non-python posts in a separate prevnext cycle.

## strategy

There are currently two supported strategies.

* first
* all

### first

`first` will cycle through only the posts contained within the first map that
contains the post.

### all

`all` will cycle through all of the posts aggregated from any prevnext map.

"""
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional

from jinja2 import Template

from markata.hookspec import hook_impl, register_attr

if TYPE_CHECKING:
    from frontmatter import Post

    from markata import Markata

SUPPORTED_STRATEGIES = ["first", "all"]


class UnsupportedPrevNextStrategy(NotImplementedError):
    """
    A custom error class to raise when an unsupporte prevnext strategy is
    defined.
    """


@dataclass
class PrevNext:
    prev: str
    next: str


def prevnext(
    markata: "Markata",
    post: "Post",
    conf: List[Dict[str, str]],
    strategy: str = "first",
) -> Optional[PrevNext]:
    posts = []
    for map_conf in conf.values():
        _posts = markata.map("post", **map_conf)
        # if the strategy is first, cycle back to the beginning after each map
        if strategy == "first" and _posts:
            _posts.append(_posts[0])
        posts.extend(_posts)
    # if the strategy is 'all', cycle back to the beginning after all of the maps.
    if strategy == "all":
        posts.append(posts[0])

    try:
        post_idx = posts.index(post)
        return PrevNext(prev=posts[post_idx - 1], next=posts[post_idx + 1])
    except ValueError:
        # post is not in posts
        return None


TEMPLATE = """
<div class='prevnext'>

    <style type='text/css'>

    :root {
      --prevnext-color-text: {{ config.get('prevnext_color_text', config.get('color_text', '#eefbfe')) }};
      --prevnext-color-angle: {{ config.get('prevnext_color_angle', config.get('color_accent', '#e1bd00c9')) }};
      --prevnext-subtitle-brightness: 3;
    }
    [data-theme="light"] {
      --prevnext-color-text: {{ config.get('prevnext_color_text_light', config.get('color_text_light', '#1f2022')) }};
      --prevnext-color-angle: {{ config.get('prevnext_color_angle_light', config.get('color_accent_light', '#ffeb00')) }};
      --prevnext-subtitle-brightness: 3;
    }
    [data-theme="dark"] {
      --prevnext-color-text: {{ config.get('prevnext_color_text', config.get('color_text', '#eefbfe')) }};
      --prevnext-color-angle: {{ config.get('prevnext_color_angle', config.get('color_accent', '#e1bd00c9')) }};
      --prevnext-subtitle-brightness: 3;
    }
    .prevnext {
      display: flex;
      flex-direction: row;
      justify-content: space-around;
      align-items: flex-start;
    }
    .prevnext a {
      display: flex;
      align-items: center;
      width: 100%;
      text-decoration: none;
    }
    a.next {
      justify-content: flex-end;
    }
    .prevnext a:hover {
      background: #00000006;
    }
    .prevnext-subtitle {
      color: var(--prevnext-color-text);
      filter: brightness(var(--prevnext-subtitle-brightness));
      font-size: .8rem;
    }
    .prevnext-title {
      color: var(--prevnext-color-text);
      font-size: 1rem;
    }
    .prevnext-text {
      max-width: 30vw;
    }
    </style>
    {% if prevnext.prev['slug'] == 'index' %}
    <a class='prev' href='/'>
    {% else %}
    <a class='prev' href='/{{ prevnext.prev['slug'] }}'>
    {% endif %}

        <svg width="50px" height="50px" viewbox="0 0 24 24" fill="none" xmlns="http://www.w3.org/2000/svg">
            <path d="M13.5 8.25L9.75 12L13.5 15.75" stroke="var(--prevnext-color-angle)" stroke-width="1.5" stroke-linecap="round" stroke-linejoin="round"> </path>
        </svg>
        <div class='prevnext-text'>
            <p class='prevnext-subtitle'>prev</p>
            <p class='prevnext-title'>{{ prevnext.prev['title'] }}</p>
        </div>
    </a>
    {% if prevnext.next['slug'] == 'index' %}
    <a class='next' href='/'>
    {% else %}
    <a class='next' href='/{{ prevnext.next['slug'] }}'>
    {% endif %}
        <div class='prevnext-text'>
            <p class='prevnext-subtitle'>next</p>
            <p class='prevnext-title'>{{ prevnext.next['title'] }}</p>
        </div>
        <svg width="50px" height="50px" viewbox="0 0 24 24" fill="none" xmlns="http://www.w3.org/2000/svg">
            <path d="M10.5 15.75L14.25 12L10.5 8.25" stroke="var(--prevnext-color-angle)" stroke-width="1.5" stroke-linecap="round" stroke-linejoin="round"></path>
        </svg>
    </a>
  </div>
"""


@hook_impl
@register_attr("prevnext")
def pre_render(markata: "Markata") -> None:

    config = markata.config.get("prevnext", {})
    feed_config = markata.config.get("feeds", {})
    strategy = config.get("strategy", "first")
    if strategy not in SUPPORTED_STRATEGIES:
        msg = f"""
        "{strategy}" is not a supported prevnext strategy

        configure prevnext in your markata.toml to use one of {SUPPORTED_STRATEGIES}
        """
        raise UnsupportedPrevNextStrategy(msg)
    template = config.get("template", None)
    if template is None:
        template = Template(TEMPLATE)
    else:
        template = Template(Path(template).read_text())
    for article in set(markata.articles):
        article["prevnext"] = prevnext(
            markata,
            article,
            feed_config,
            strategy=strategy,
        )
        if "prevnext" not in article.content and article["prevnext"]:
            article.content += template.render(config=config, **article)



"""
Writes the final modified markdown and frontmatter to the output directory.
Replacing the trailing slash if its there and adding .md will bring up the raw
source.

## Configuration

The only configuration for the publish_source plugin is to make sure its in
your list of hooks.


``` toml
[markata]

# make sure its in your list of hooks
hooks=[
   "markata.plugins.publish_source",
   ]

!! note
    publish_source is included by default, but if you have not included the
    default set of hooks you will need to explicitly add it.
"""
from pathlib import Path
from typing import TYPE_CHECKING

import frontmatter
import yaml
from yaml.representer import RepresenterError

from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata import Markata


def _save(output_dir: Path, article: frontmatter.Post) -> None:
    """
    saves the article to the output directory at its specified slug.
    """
    with open(
        output_dir / Path(article["slug"]).parent / Path(article["path"]).name, "w+"
    ) as f:
        f.write(frontmatter.dumps(article))


def _strip_unserializable_values(article: frontmatter.Post) -> frontmatter.Post:
    """
    Returns an article with only yaml serializable frontmatter.
    """
    _article = frontmatter.Post(
        article.content, **{k: v for k, v in article.metadata.items() if k != "content"}
    )
    kwargs = {
        "Dumper": yaml.cyaml.CSafeDumper,
        "default_flow_style": False,
        "allow_unicode": True,
    }
    for key, value in article.metadata.items():
        try:
            yaml.dump({key: value}, **kwargs)
        except RepresenterError:
            del _article[key]
    return _article


@hook_impl
def save(markata: "Markata") -> None:
    """
    Saves the final modified post to the output site as markdown.

    !! note
      Any keys that are not yaml serializable will be stripped.
    """
    output_dir = Path(str(markata.config["output_dir"]))
    output_dir.mkdir(parents=True, exist_ok=True)
    for article in markata.iter_articles(description="saving source documents"):
        try:
            _save(output_dir, article)
        except RepresenterError:
            _article = _strip_unserializable_values(article)

            _save(output_dir, _article)



from .cli import app, cli, make_layout, run_until_keyboard_interrupt
from .header import Header
from .plugins import Plugins
from .runner import Runner
from .server import Server
from .summary import Summary

__all__ = [
    "Header",
    "Plugins",
    "Runner",
    "Server",
    "Summary",
    "app",
    "cli",
    "make_layout",
    "run_until_keyboard_interrupt",
]



from rich.console import RenderableType
from textual.app import App
from textual.widget import Widget
from textual.widgets import Footer

from markata import Markata
from markata.hookspec import hook_impl


class MarkataWidget(Widget):
    def __init__(self, markata: Markata, widget: str = "server"):
        super().__init__(widget)
        self.m = markata
        self.widget = widget
        self.renderable = getattr(self.m, self.widget)

    def render(self):
        return self.renderable

    async def update(self, renderable: RenderableType) -> None:
        self.renderable = renderable
        self.refresh()


class MarkataApp(App):
    async def on_mount(self) -> None:
        self.m = Markata()
        self.m.console.quiet = True
        self.server = MarkataWidget(self.m, "server")
        self.runner = MarkataWidget(self.m, "runner")
        self.plugins = MarkataWidget(self.m, "plugins")
        self.summary = MarkataWidget(self.m, "summary")
        await self.view.dock(Footer(), edge="bottom")
        await self.view.dock(self.plugins, edge="left", size=30, name="plugins")
        await self.view.dock(self.summary, edge="right", size=30, name="summary")
        await self.view.dock(self.server, self.runner, edge="top")
        self.set_interval(1, self.action_refresh)

    async def on_load(self, event):
        await self.bind("q", "quit", "quit")
        await self.bind("r", "refresh", "refresh")

    async def action_refresh(self) -> None:
        self.refresh()
        self.runner.refresh()
        self.server.refresh()
        self.plugins.refresh()
        self.summary.refresh()


@hook_impl()
def cli(app, markata):
    @app.command()
    def tui():
        MarkataApp.run(log="textual.log")


if __name__ == "__main__":
    MarkataApp.run(log="textual.log")



"""Standard Config.
A module to load tooling config from a users project space.

Inspired from frustrations that some tools have a tool.ini, .tool.ini,
setup.cfg, or pyproject.toml.  Some allow for global configs, some don't.  Some
properly follow the users home directory, others end up in a weird temp
directory.  Windows home directory is only more confusing.  Some will even
respect the users `$XDG_HOME` directory.


This file is for any project that can be configured in plain text such as `ini`
or `toml` and not requiring a .py file.  Just name your tool and let users put
config where it makes sense to them, no need to figure out resolution order.

## Usage:

``` python
from standard_config import load

# Retrieve any overrides from the user
overrides = {'setting': True}
config = load('my_tool', overrides)
```

## Resolution Order

* First global file with a tool key
* First local file with a tool key
* Environment variables prefixed with `TOOL`
* Overrides

### Tool Specific Ini files

Ini file formats must include a `<tool>` key.

``` ini
[my_tool]
setting = True
```

### pyproject.toml

Toml files must include a `tool.<tool>` key

``` toml
[tool.my_tool]
setting = True
```

### setup.cfg

setup.cfg files must include a `tool:<tool>` key

``` ini
[tool:my_tool]
setting = True
```


### global files to consider

* <home>/tool.ini
* <home>/.tool
* <home>/.tool.ini
* <home>/.config/tool.ini
* <home>/.config/.tool
* <home>/.config/.tool.ini

### local files to consider

* <project_home>/tool.ini
* <project_home>/.tool
* <project_home>/.tool.ini
* <project_home>/pyproject.toml
* <project_home>/setup.cfg

"""

import os
from pathlib import Path
from typing import Dict, List, Union

import anyconfig

# path_spec_type = List[Dict[str, Union[Path, str, List[str]]]]
path_spec_type = List


def _get_global_path_specs(tool: str) -> path_spec_type:
    """
    Generate a list of standard pathspecs for global config files.

    Args:
        tool (str): name of the tool to configure
    """
    try:
        home = Path(os.environ["XDG_HOME"])
    except KeyError:
        home = Path.home()

    return [
        {"path_specs": home / f"{tool}.ini", "ac_parser": "ini", "keys": [tool]},
        {"path_specs": home / f".{tool}", "ac_parser": "ini", "keys": [tool]},
        {"path_specs": home / f".{tool}.ini", "ac_parser": "ini", "keys": [tool]},
        {
            "path_specs": home / ".config" / f"{tool}.ini",
            "ac_parser": "ini",
            "keys": [tool],
        },
        {
            "path_specs": home / ".config" / f".{tool}",
            "ac_parser": "ini",
            "keys": [tool],
        },
        {
            "path_specs": home / ".config" / f".{tool}.ini",
            "ac_parser": "ini",
            "keys": [tool],
        },
    ]


def _get_local_path_specs(tool: str, project_home: Union[str, Path]) -> path_spec_type:
    """
    Generate a list of standard pathspecs for local, project directory config files.

    Args:
        tool (str): name of the tool to configure
    """
    return [
        {
            "path_specs": Path(project_home) / f"{tool}.ini",
            "ac_parser": "ini",
            "keys": [tool],
        },
        {
            "path_specs": Path(project_home) / f".{tool}",
            "ac_parser": "ini",
            "keys": [tool],
        },
        {
            "path_specs": Path(project_home) / f".{tool}.ini",
            "ac_parser": "ini",
            "keys": [tool],
        },
        {
            "path_specs": Path(project_home) / f"{tool}.yml",
            "ac_parser": "yaml",
            "keys": [tool],
        },
        {
            "path_specs": Path(project_home) / f".{tool}.yml",
            "ac_parser": "yaml",
            "keys": [tool],
        },
        {
            "path_specs": Path(project_home) / f"{tool}.toml",
            "ac_parser": "toml",
            "keys": [tool],
        },
        {
            "path_specs": Path(project_home) / f".{tool}.toml",
            "ac_parser": "toml",
            "keys": [tool],
        },
        {
            "path_specs": Path(project_home) / f"pyproject.toml",
            "ac_parser": "toml",
            "keys": ["tool", tool],
        },
        {
            "path_specs": Path(project_home) / f"setup.cfg",
            "ac_parser": "ini",
            "keys": [f"tool.{tool}"],
        },
    ]


def _get_attrs(attrs: list, config: Dict) -> Dict:
    """Get nested config data from a list of keys.

    specifically written for pyproject.toml which needs to get `tool` then `<tool>`
    """
    for attr in attrs:
        config = config[attr]
    return config


def _load_files(config_path_specs: path_spec_type) -> Dict:
    """Use anyconfig to load config files stopping at the first one that exists.

    config_path_specs (list): a list of pathspecs and keys to load
    """
    for file in config_path_specs:

        if file["path_specs"].exists():
            config = anyconfig.load(**file)
        else:
            # ignore missing files
            continue

        try:
            return _get_attrs(file["keys"], config)
        except KeyError:
            # ignore incorrect keys
            continue

    return {}


def _load_env(tool: str) -> Dict:
    """Load config from environment variables.

    Args:
        tool (str): name of the tool to configure
    """
    vars = [var for var in os.environ.keys() if var.startswith(tool.upper())]
    return {
        var.lower().strip(tool.lower()).strip("_").strip("-"): os.environ[var]
        for var in vars
    }


def load(tool: str, project_home: Union[Path, str] = ".", overrides: Dict = {}) -> Dict:
    """Load tool config from standard config files.

    Resolution Order

    * First global file with a tool key
    * First local file with a tool key
    * Environment variables prefixed with `TOOL`
    * Overrides

    Args:
        tool (str): name of the tool to configure
    """
    global_config = _load_files(_get_global_path_specs(tool))
    local_config = _load_files(_get_local_path_specs(tool, project_home))
    env_config = _load_env(tool)
    return {**global_config, **local_config, **env_config, **overrides}



from pathlib import Path

import anyconfig

from markata import Markata
from markata.hookspec import hook_impl, register_attr


@hook_impl
@register_attr("sitemap")
def render(markata: Markata) -> None:
    url = markata.get_config("url") or ""

    sitemap = {
        "urlset": [
            {
                "url": {
                    "loc": url + "/" + article["slug"] + "/",
                    "changefreq": "daily",
                    "priority": "0.7",
                }
            }
            for article in markata.articles
            if article["status"] == "published"
        ]
    }

    sitemap = (
        anyconfig.dumps(sitemap, "xml")
        .decode("utf-8")
        .replace(
            "<urlset>",
            '<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9" xmlns:news="http://www.google.com/schemas/sitemap-news/0.9" xmlns:xhtml="http://www.w3.org/1999/xhtml" xmlns:mobile="http://www.google.com/schemas/sitemap-mobile/1.0" xmlns:image="http://www.google.com/schemas/sitemap-image/1.1" xmlns:video="http://www.google.com/schemas/sitemap-video/1.1">',
        )
        .replace("</url>", "</url>\n")
    )
    setattr(markata, "sitemap", sitemap)


@hook_impl
def save(markata: Markata) -> None:
    with open(Path(markata.config["output_dir"]) / "sitemap.xml", "w") as f:
        f.write(markata.sitemap)  # type: ignore



"""add generator meta tag"""
from bs4 import BeautifulSoup

from markata import Markata, __version__
from markata.hookspec import hook_impl


@hook_impl(trylast=True)
def render(markata: Markata) -> None:
    for article in markata.iter_articles("add ssg tag"):
        soup = BeautifulSoup(article.html, features="lxml")
        tag = soup.new_tag("meta")
        tag.attrs["content"] = f"markata {__version__}"
        tag.attrs["name"] = "generator"
        soup.head.append(tag)
        article.soup = soup
        article.html = soup.prettify()



"""
Sets the articles `output_html` path, and saves the article's `html` to the
`output_html` file.

## Ouptut Directory

Output will always be written inside of the configured `output_dir`

```toml
[markata]
# markout is the default, but you can override it in your markata.toml file
output_dir = "markout"
```

## Explicityly set the output

markata will save the articles `html` to the `output_html` specified in the
articles metadata, loaded from frontmatter.

### 404 example use case

Here is an example use case of explicitly setting the output_html.  By default
markata will turn `pages/404.md` into `markout/404/index.html`, but many
hosting providers look for a 404.html to redirect the user to when a page is
not found.

```markdown
---
title: Whoops that page was not found
description: 404, looks like we can't find the page you are looking for
output_html: 404.html

---

404, looks like we can't find the page you are looking for.  Try one of these
pages.

<ul>
{% for post in markata.map('post', filter='"markata" not in slug and "tests" not in slug and "404" not in slug') %}
    <li><a href="{{ post.slug }}">{{ post.title or "CHANGELOG" }}</a></li>
{% endfor %}
</ul>
```

## Index.md is the one special case

If you have a file `pages/index.md` it will become `markout/index.html` rather
than `markout/index/inject.html` This is one of the primary ways that markata
lets you [make your home page](https://markata.dev/home-page/)

"""
from pathlib import Path
from typing import TYPE_CHECKING

from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata import Markata


def _is_relative_to(output_dir: Path, output_html: Path):
    try:
        output_html.relative_to(output_dir)
        return True
    except ValueError:
        return False


@hook_impl
def pre_render(markata: "Markata") -> None:
    """
    Sets the `output_html` in the articles metadata.  If the output is
    explicitly given, it will make sure its in the `output_dir`, if it is not
    explicitly set it will use the articles slug.
    """
    output_dir = Path(markata.config["output_dir"])  # type: ignore
    output_dir.mkdir(parents=True, exist_ok=True)

    for article in markata.articles:
        if "output_html" in article.metadata:
            article_path = Path(article["output_html"])
            if not _is_relative_to(output_dir, article_path):
                article["output_html"] = output_dir / article["output_html"]
        elif article["slug"] == "index":
            article["output_html"] = output_dir / "index.html"
        else:
            article["output_html"] = output_dir / article["slug"] / "index.html"


@hook_impl
def save(markata: "Markata") -> None:
    """
    Saves all the articles to their set `output_html` location if that location
    is relative to the specified `output_dir`.  If its not relative to the
    `output_dir` it will log an error and move on.
    """
    output_dir = Path(markata.config["output_dir"])  # type: ignore

    for article in markata.articles:
        article_path = Path(article["output_html"])
        if _is_relative_to(output_dir, article_path):
            article_path.parent.mkdir(parents=True, exist_ok=True)
            with open(article_path, "w+") as f:
                f.write(article.html)
        else:
            markata.console.log(
                f'article "{article["path"]}" attempted to write to "{article["output_html"]}" outside of the configured output_dir "{output_dir}"'
            )



"""
The `markata.plugins.feeds` plugin is used to create feed pages, which are lists of
posts.  The list is generated using a `filter`, then each post in the list is
rendered with a `card_template` before being applied to the `body` of the
`template`.

## Installation

This plugin is built-in and enabled by default, but in you want to be very
explicit you can add it to your list of existing plugins.

``` toml
hooks = [
   "markata.plugins.feeds",
   ]
```

## Configuration

### set default template and card_template

At the root of the markata.feeds config you may set `template`, and
`card_template`.  These will become your defaults for every feed you create.
If you do not set these, markata will use it's defaults.  The defaults are
designed to work for a variety of use cases, but are not likely the best for
all.

``` toml
[markata.feeds_config]
template="pages/templates/archive_template.html"
card_template="plugins/feed_card_template.html"
```

### pages

Underneath of the `markata.feeds` we will create a new map for each page where
the name of the map will be the name of the page.


The following config will create a page at `/all-posts` that inclues every
single post.

``` toml
[markata.feeds.all-posts]
filter="True"
```

### template

The `template` configuration key is a file path to the template that you want
to use to create the feed.  You may set the default template you want to use
for all feeds under `[markata.feeds]`, as well as override it inside of each
feeds config.

The template is a jinja style template that expects to fill in a `title` and
`body` variable.

``` html
<!DOCTYPE html>
<html lang="en">
  <head>
    <title>{{ title }}</title>
  </head>
  <body>
    <ul>
        {{ body }}
    </ul>
  </body>
</html>
```

!!! note
    I highly reccomend putting your `body` in a `<ul>`, and wrapping your
    `card_template`s in an `<li>`.

### card_template

All keys available from each post is available to put into your jinja
template.  These can either be placed there in your post frontmatter, or
through a plugin that automatically adds to the post before the save phase.

Here is a very simple example that would give a link to each post with the
title and date.

``` toml
[markata.feeds.all]
filter="True"
card_template='<li><a href={{slug}}>{{title}}-{{date}}</a></li>'
```

### filter

The filter is a python expression ran on every post that expects to return a
boolean.  The variables available to this expression are every key in your
frontmatter, plus the `timedelta` function, and `parse` function to more easily
work with dates.

## Feed Examples

True can be passed in to make a feed of all the posts you have.

``` toml
[markata.feeds.archive]
filter="True"
```

You can compare against the values of the keys from your frontmatter.  This
example creates a feed that includes every post with the status of `draft`.

``` toml
[markata.feeds.draft]
filter="status=='draft'"
```

We can also compare against dates.  The
[markata.plugins.datetime](https://markata.dev/markata/plugins/datetime/)
plugin, automatically adds `today` as today's date and `now` as the current
datetime.  These are quite handy to create feeds for scheduled, recent, or
today's posts.  The following two examples will create a feed for scheduled
posts and for today's posts respectively.

``` toml
[markata.feeds.scheduled]
filter="date>today"

[markata.feeds.today]
filter="date==today"
```

If you have list of items in your frontmatter for something like `tags`, you
can check for the existence of a tag in the list.

``` toml
[markata.feeds.python]
filter="date<=today and 'python' in tags"
```

And of course you can combine all the things into larger expressions.  Here is one example of the main feed on my blog.

``` toml
[markata.feeds.blog]
filter="date<=today and templateKey in ['blog-post'] and status.lower()=='published'"
```

Here is another example that shows my drafts for a particular tag.

``` toml
[markata.feeds.python-draft]
filter="date<=today and 'python' in tags and status=='draft'"
```

## Defaults

By default feeds will create one feed page at `/archive/` that includes all
posts.

[markata.feeds.archive]
filter="True"

"""
import datetime
import shutil
import textwrap
from pathlib import Path
from typing import TYPE_CHECKING, Any, List, Optional, Union

from jinja2 import Template, Undefined

from markata import Markata
from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from frontmatter import Post


class SilentUndefined(Undefined):
    def _fail_with_undefined_error(self, *args, **kwargs):
        return ""


class MarkataFilterError(RuntimeError):
    ...


@hook_impl
def configure(markata: Markata) -> None:
    """
    configure the default values for the feeds plugin
    """
    config = markata.config.get("feeds", {})
    if config is None:
        markata.config["feeds"] = dict()
        config = markata.config.get("feeds", {})
    if "archive" not in config.keys():
        config["archive"] = dict()
        config["archive"]["filter"] = "True"

    default_post_template = config.get(
        "template", Path(__file__).parent / "default_post_template.html"
    )
    for page, page_conf in config.items():
        if "template" not in page_conf.keys():
            page_conf["template"] = default_post_template


@hook_impl
def save(markata: Markata) -> None:
    """
    Creates a new feed page for each page in the config.
    """
    feeds = markata.config.get("feeds", {})

    description = markata.get_config("description") or ""
    url = markata.get_config("url") or ""

    for page, page_conf in feeds.items():

        create_page(
            markata,
            page,
            description=description,
            url=url,
            **page_conf,
        )

    home = Path(str(markata.config["output_dir"])) / "index.html"
    archive = Path(str(markata.config["output_dir"])) / "archive" / "index.html"
    if not home.exists() and archive.exists():
        shutil.copy(str(archive), str(home))


def create_page(
    markata: Markata,
    page: str,
    tags: Optional[List] = None,
    status: str = "published",
    template: Optional[Union[Path, str]] = None,
    card_template: Optional[str] = None,
    filter: Optional[str] = None,
    description: Optional[str] = None,
    url: Optional[str] = None,
    title: Optional[str] = "feed",
    sort: str = "True",
    reverse: bool = False,
) -> None:
    """
    create an html unorderd list of posts.
    """

    posts = markata.map("post", filter=filter, sort=sort, reverse=reverse)
    # if filter is not None:
    #     posts = reversed(
    #         sorted(
    #             markata.articles, key=lambda x: x.get("date", datetime.date(1970, 1, 1))
    #         )
    #     )
    #     try:
    #         posts = [post for post in posts if eval(filter, post.to_dict(), {})]
    #     except Exception as e:
    #         msg = textwrap.dedent(
    #             f"""
    #                 While processing feed page='{page}' markata hit the following exception
    #                 during filter='{filter}'
    #                 {e}
    #                 """
    #         )
    #         raise MarkataFilterError(msg)

    cards = [create_card(post, card_template) for post in posts]
    cards.insert(0, "<ul>")
    cards.append("</ul>")

    with open(template) as f:
        template = Template(f.read(), undefined=SilentUndefined)
    output_file = Path(markata.config["output_dir"]) / page / "index.html"
    canonical_url = f"{url}/{page}/"
    output_file.parent.mkdir(exist_ok=True, parents=True)

    with open(output_file, "w+") as f:
        f.write(
            template.render(
                body="".join(cards),
                url=url,
                description=description,
                title=title,
                canonical_url=canonical_url,
                today=datetime.datetime.today(),
                config=markata.config,
            )
        )


def create_card(post: "Post", template: Optional[str] = None) -> Any:
    """
    Creates a card for one post based on the configured template.  If no
    template is configured it will create one with the post title and dates (if present).
    """
    if template is None:
        if "date" in post.keys():
            return textwrap.dedent(
                f"""
                <li class='post'>
                <a href="/{post['slug']}/">
                    {post['title']} {post['date'].year}-{post['date'].month}-{post['date'].day}
                </a>
                </li>
                """
            )
        else:
            return textwrap.dedent(
                f"""
                <li class='post'>
                <a href="/{post['slug']}/">
                    {post['title']}
                </a>
                </li>
                """
            )
    try:
        _template = Template(Path(template).read_text())
    except FileNotFoundError:
        _template = Template(template)
    return _template.render(**post.to_dict())



"""
markata uses setup tools for packaging.

To Build markata as a Python package

    $ python setup.py sdist bdist_wheel --bdist-dir ~/temp/bdistwheel

Regular install

    $ pip install -e .

To setup local Development

    $ pip install -e .

"""
from pathlib import Path

from setuptools import find_packages, setup

NAME = "markata"

README = (Path(__file__).parent / "README.md").read_text(encoding="utf-8")

with open("requirements.txt", "r", encoding="utf-8") as f:
    requires = [x.strip() for x in f if x.strip()]

with open("requirements_dev.txt", "r", encoding="utf-8") as f:
    dev_requires = [x.strip() for x in f if x.strip()]

README = (Path(__file__).parent / "README.md").read_text(encoding="utf-8")

setup(
    name=NAME,
    version="0.2.0",
    url="https://markata.dev",
    author="Waylon Walker",
    author_email="waylon@waylonwalker.com",
    description="Static site generator plugins all the way down.",
    long_description=README,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    entry_points={"console_scripts": ["markata=markata.cli:cli"]},
    platforms="any",
    license="MIT",
    install_requires=requires,
    extras_require={"dev": dev_requires},
    zip_safe=False,
    include_package_data=True,
)



"""
Renders your markdown as a jinja template during pre_render.

The markata instance is passed into the template, giving you access to things
such as all of your articles, config, and this post as post.

## Examples

first we can grab a few things out of the frontmatter of this post.

``` markdown
### {{ post.title }}
{{ post.description }}
```

### one-liner list of links

This one-liner will render a list of markdown links into your markdown at build
time.  It's quite handy to pop into posts.

``` markdown
{{ '\\n'.join(markata.map('f"* [{title}]({slug})"', sort='slug')) }}
```

### jinja for to markdown list of links

Sometimes quoting things like your filters are hard to do in a one line without
running out of quote variants.  Jinja for loops can make this much easier.

``` markdown
{% for post in markata.map('post', filter='"git" in tags') %}
* [{{ post.title }}]({{ post.slug }})
{% endfor %}
```

### jinja for to html list of links

Since markdown is a superset of html, you can just render out html into your
post and it is still valid.

``` markdown
<ul>
{% for post in markata.map('post', filter='"git" in tags') %}
    <li><a href="{{ post.slug }}">{{ post.title }}</a></li>
{% endfor %}
</ul>
```

## Ignoring files

It is possible to ignore files by adding an ignore to your `markata.jinja_md`
config in your `markata.toml` file.  This ignore follows the `gitwildmatch`
rules, so think of it the same as writing a gitignore.

```toml
[markata.jinja_md]
ignore=[
'jinja_md.md',
]
```

!!note
  Docs such as this jinja_md.py file will get converted to jinja_md.md during
  build time, so use `.md` extensions instead of `.py`.

## Ignoring a single file

You can also ignore a single file right from the articles frontmatter, by adding `jinja: false`.

```markdown
---
jinja: false

---
```

## Escaping

Sometimes you want the ability to have jinja templates in a post, but also the
ability to keep a raw jinja template.  There are a couple of techniques that
are covered mroe in the jinja docs for
[escaping](https://jinja.palletsprojects.com/en/3.1.x/templates/#escaping)

```markdown
{% raw %}
{{ '\\n'.join(markata.map('f"* [{title}]({slug})"', sort='slug')) }}
{% endraw %}

{{ '{{' }} '\\n'.join(markata.map('f"* [{title}]({slug})"', sort='slug')) {{ '}}' }}
```
"""
from pathlib import Path
from typing import TYPE_CHECKING

import jinja2
import pathspec
from jinja2 import TemplateSyntaxError, Undefined, UndefinedError, nodes
from jinja2.ext import Extension

from markata.hookspec import hook_impl, register_attr


class IncludeRawExtension(Extension):
    tags = {"include_raw"}

    def parse(self, parser):
        line_number = next(parser.stream).lineno
        file = [parser.parse_expression()]
        return nodes.CallBlock(
            self.call_method("_read_file", file), [], [], ""
        ).set_lineno(line_number)

    def _read_file(self, file, caller):
        return Path(file).read_text()


if TYPE_CHECKING:
    from markata import Markata


class _SilentUndefined(Undefined):
    """
    silence undefined variable errors in jinja templates.

    ### Example
    ```python
    template = '{{ variable }}'
    article.content = Template( template, undefined=_SilentUndefined).render()
    ```
    """

    def _fail_with_undefined_error(self, *args, **kwargs):
        return ""


class PostTemplateSyntaxError(TemplateSyntaxError):
    """
    Custom error message for post template syntax errors.
    """


@hook_impl
@register_attr("prevnext")
def pre_render(markata: "Markata") -> None:
    """
    jinja_md hook for markata to render your markdown post as a jinja template.

    The post itself is exposed as `post`, and the markata instance is exposed
    as `markata`.
    """

    config = markata.get_plugin_config("jinja_md")
    ignore_spec = pathspec.PathSpec.from_lines("gitwildmatch", config.get("ignore", []))
    # for article in markata.iter_articles(description="jinja_md"):
    jinja_env = jinja2.Environment(
        extensions=[
            IncludeRawExtension,
        ],
    )
    for article in markata.articles:
        if article.get("jinja", True) and not ignore_spec.match_file(article["path"]):
            try:

                article.content = jinja_env.from_string(article.content).render(
                    markata=markata, post=article
                )
                # prevent double rendering
                article["jinja"] = False
            except TemplateSyntaxError as e:
                errorline = article.content.split("\n")[e.lineno - 1]
                msg = f"""
                Error while processing post {article['path']}

                {errorline}
                """

                raise PostTemplateSyntaxError(msg, lineno=e.lineno)
            except UndefinedError:
                ...



import json
from pathlib import Path
from typing import TYPE_CHECKING

from markata.hookspec import hook_impl

if TYPE_CHECKING:
    from markata import Markata


@hook_impl
def save(markata: "Markata") -> None:
    output_file = Path(markata.config["output_dir"]) / "markata.json"
    output_file.parent.mkdir(parents=True, exist_ok=True)
    output_file.write_text(json.dumps(markata.to_dict(), default=str))



import atexit
import time
from pathlib import Path
from typing import Union

from rich.panel import Panel


def find_port(port: int = 8000) -> int:
    """Find a port not in ues starting at given port"""
    import socket

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        if s.connect_ex(("localhost", port)) == 0:
            return find_port(port=port + 1)
        else:
            return port


class Server:
    def __init__(
        self,
        auto_restart: bool = True,
        directory: Union[str, "Path"] = None,
        port: int = 8000,
    ):
        if directory is None:
            from markata import Markata

            m = Markata()
            directory = Path(str(m.config["output_dir"]))

        self.auto_restart = auto_restart
        self.directory = directory
        self.port = find_port(port=port)
        self.start_server()
        atexit.register(self.kill)

    def start_server(self) -> None:
        import subprocess

        self.cmd = [
            "python",
            "-m",
            "http.server",
            str(self.port),
            "--directory",
            self.directory,
        ]

        self.proc = subprocess.Popen(
            self.cmd,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        self.start_time = time.time()

    @property
    def uptime(self) -> int:
        return round(time.time() - self.start_time)

    def kill(self) -> None:
        self.auto_restart = False
        self.proc.kill()

    def __rich__(self) -> Panel:
        if not self.proc.poll():
            return Panel(
                f"[green]serving on port: [gold1]{self.port} [green]using pid: [gold1]{self.proc.pid} [green]uptime: [gold1]{self.uptime} [green]link: [gold1] http://localhost:{self.port}[/]",
                border_style="blue",
                title="server",
            )

        else:

            return Panel(f"[red]server died", title="server", border_style="red")


if __name__ == "__main__":

    from rich.live import Live

    from markata import Markata

    from .cli import run_until_keyboard_interrupt

    m = Markata()
    with Live(Server(), refresh_per_second=1, screen=True):
        run_until_keyboard_interrupt()



"""
Creates links next to all heading tags to make it easier for users to share a
specific heading.
"""
import re
from pathlib import Path
from typing import TYPE_CHECKING, Any

from bs4 import BeautifulSoup

from markata import Markata
from markata.hookspec import hook_impl

if TYPE_CHECKING:
    import bs4
    from frontmatter import Post


@hook_impl(trylast=True)
def post_render(markata: Markata) -> None:
    """
    This plugin creates a link svg next to all headings.
    """

    config = markata.get_plugin_config(__file__)
    with markata.cache as cache:
        for article in markata.iter_articles("link headers"):

            key = markata.make_hash(
                "heading_link",
                "post_render",
                Path(__file__).read_text(),
                article.content,
                article.html,
            )

            html_from_cache = cache.get(key)

            if html_from_cache is None:
                html = link_headings(article)
                cache.add(key, html, expire=config["cache_expire"])
            else:
                html = html_from_cache
            article.html = html


def link_headings(article: "Post") -> Any:
    """
    Use BeautifulSoup to find all headings and run link_heading on them.
    """
    soup = BeautifulSoup(article.html, "html.parser")
    for heading in soup.find_all(re.compile("^h[1-6]$")):
        if (
            not heading.find("a", {"class": "heading-permalink"})
            and heading.get("id", "") != "title"
        ):
            link_heading(soup, heading)
    html = soup.prettify()
    return html


def link_heading(soup: "bs4.BeautifulSoup", heading: "bs4.element.Tag") -> None:
    """
    Mutate soup to include an svg link at the heading passed in.
    """
    id = heading.get("id")

    link = soup.new_tag(
        "a",
        alt="id",
        title=f"link to #{id}",
        href=f"#{id}",
        **{"class": "heading-permalink"},
    )
    span = soup.new_tag("span", **{"class": "visually-hidden"})
    svg = soup.new_tag(
        "svg",
        fill="currentColor",
        focusable="false",
        width="1em",
        height="1em",
        xmlns="http://www.w3.org/2000/svg",
        viewBox="0 0 24 24",
        **{
            "aria-hidden": "true",
        },
    )

    path = soup.new_tag(
        "path",
        d="M9.199 13.599a5.99 5.99 0 0 0 3.949 2.345 5.987 5.987 0 0 0 5.105-1.702l2.995-2.994a5.992 5.992 0 0 0 1.695-4.285 5.976 5.976 0 0 0-1.831-4.211 5.99 5.99 0 0 0-6.431-1.242 6.003 6.003 0 0 0-1.905 1.24l-1.731 1.721a.999.999 0 1 0 1.41 1.418l1.709-1.699a3.985 3.985 0 0 1 2.761-1.123 3.975 3.975 0 0 1 2.799 1.122 3.997 3.997 0 0 1 .111 5.644l-3.005 3.006a3.982 3.982 0 0 1-3.395 1.126 3.987 3.987 0 0 1-2.632-1.563A1 1 0 0 0 9.201 13.6zm5.602-3.198a5.99 5.99 0 0 0-3.949-2.345 5.987 5.987 0 0 0-5.105 1.702l-2.995 2.994a5.992 5.992 0 0 0-1.695 4.285 5.976 5.976 0 0 0 1.831 4.211 5.99 5.99 0 0 0 6.431 1.242 6.003 6.003 0 0 0 1.905-1.24l1.723-1.723a.999.999 0 1 0-1.414-1.414L9.836 19.81a3.985 3.985 0 0 1-2.761 1.123 3.975 3.975 0 0 1-2.799-1.122 3.997 3.997 0 0 1-.111-5.644l3.005-3.006a3.982 3.982 0 0 1 3.395-1.126 3.987 3.987 0 0 1 2.632 1.563 1 1 0 0 0 1.602-1.198z",
    )
    svg.append(path)
    link.append(span)
    link.append(svg)
    heading.append(link)



from typing import TYPE_CHECKING, Union

from rich.panel import Panel
from rich.table import Table

if TYPE_CHECKING:
    from markata import Markata


class Summary:
    def __init__(self, m: "Markata", simple: bool = False) -> None:
        self.m = m
        self.simple = simple

    def __rich__(self) -> Union[Panel, Table]:
        grid = Table.grid(expand=True)
        grid.add_row(f"[bright_blue]{len(self.m.articles)}[/] articles")
        grid.add_row(
            f"[green]{len([a for a in self.m.articles if a['status'] =='published'])}[/] published"
        )
        grid.add_row(
            f"[gold1]{len([a for a in self.m.articles if a['status'] =='draft'])}[/] drafts"
        )
        grid.add_row("")
        grid.add_row("[bold gold1]TAGS[/]")
        from collections import Counter

        from more_itertools import flatten

        try:
            for tag, count in Counter(
                list(flatten([a["tags"] for a in self.m.articles]))
            ).most_common():
                grid.add_row(f'{count} {" "*(3-len(str(count)))} {tag}')
        except KeyError:
            ...

        try:
            grid.add_row("[bold gold1]Series[/]")
            for series, count in Counter(
                [a["templateKey"] for a in self.m.articles]
            ).most_common():
                grid.add_row(f'{count} {" "*(3-len(str(count)))} {series}')
        except KeyError:
            ...
        if self.simple:
            return grid
        else:
            return Panel(grid, title="[gold1]summary[/]", border_style="magenta")


if __name__ == "__main__":
    from rich import print

    m = Markata()
    print(Summary(m, simple=True))



from typing import TYPE_CHECKING

import rich
from rich.panel import Panel
from rich.table import Table

if TYPE_CHECKING:
    from markata import Markata


class Plugins:
    def __init__(self, markata: "Markata"):
        self.m = markata

    def __rich__(self) -> Panel:
        grid = Table.grid(expand=True)
        grid.add_row(f"[bright_blue]{len(self.m._pm.get_plugins())}[/] plugins")
        for plugin in self.m._pm.get_plugins():
            grid.add_row(
                "".join(
                    [
                        "[bright_black]",
                        ".".join(plugin.__name__.split(".")[:-1]),
                        ".[/]",
                        plugin.__name__.split(".")[-1],
                    ]
                )
            )
        return Panel(grid, title="plugins", border_style="gold1")


if __name__ == "__main__":

    plugins = Plugins(Markata())
    rich.print(plugins)



from pathlib import Path

from markata.hookspec import hook_impl, register_attr


@hook_impl
@register_attr("articles")
def pre_render(markata) -> None:
    for article in markata.filter('title==""'):
        article["title"] = (
            Path(article["path"]).stem.replace("-", " ").replace("_", " ").title()
        )
